# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import datetime as dt
import json
import logging
from functools import partial
from itertools import zip_longest
from os import environ, makedirs
from os.path import dirname, exists, expanduser, expandvars, isdir, join, normpath
from subprocess import call

import xdg
from click import Choice, UsageError, confirm, prompt

from .exceptions import FatalError
from .settings import find_configuration_file, utils

logger = logging.getLogger("khal")


def compressuser(path):
    """Abbreviate home directory to '~', for presenting a path."""
    home = normpath(expanduser("~"))
    path = normpath(path)
    if path.startswith(home):
        path = "~" + path[len(home) :]
    return path


def validate_int(input, min_value, max_value):
    try:
        number = int(input)
    except ValueError:
        raise UsageError("Input must be an integer")
    if min_value <= number <= max_value:
        return number
    else:
        raise UsageError(f"Input must be between {min_value} and {max_value}")


DATE_FORMAT_INFO = [
    ("Year", ["%Y", "%y"]),
    ("Month", ["%m", "%B", "%b"]),
    ("Day", ["%d", "%a", "%A"]),
]


def present_date_format_info(example_date):
    columns = []
    widths = []
    for title, formats in DATE_FORMAT_INFO:
        newcol = [title]
        for f in formats:
            newcol.append(f"{f}={example_date.strftime(f)}")
        widths.append(max(len(s) for s in newcol) + 2)
        columns.append(newcol)

    print("Common fields for date formatting:")
    for row in zip_longest(*columns, fillvalue=""):
        print("".join(s.ljust(w) for (s, w) in zip(row, widths)))

    print(
        "More info: https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior"
    )


def choose_datetime_format():
    """query user for their date format of choice"""
    choices = [
        ("year-month-day", "%Y-%m-%d"),
        ("day/month/year", "%d/%m/%Y"),
        ("month/day/year", "%m/%d/%Y"),
    ]
    validate = partial(validate_int, min_value=0, max_value=3)
    today = dt.date.today()
    print("What ordering of year, month, date do you want to use?")
    for num, (desc, fmt) in enumerate(choices):
        print(f"[{num}] {desc} (today: {today.strftime(fmt)})")
    print("[3] Custom")
    choice_no = prompt("Please choose one of the above options", value_proc=validate)
    if choice_no == 3:
        present_date_format_info(today)
        dateformat = prompt("Make your date format")
    else:
        dateformat = choices[choice_no][1]
    print(f"Date format: {dateformat} (today as an example: {today.strftime(dateformat)})")
    return dateformat


def choose_time_format():
    """query user for their time format of choice"""
    choices = ["%H:%M", "%I:%M %p"]
    print("What timeformat do you want to use?")
    print("[0] 24 hour clock (recommended)\n[1] 12 hour clock")
    validate = partial(validate_int, min_value=0, max_value=1)
    prompt_text = "Please choose one of the above options"
    timeformat = choices[prompt(prompt_text, default=0, value_proc=validate)]
    now = dt.datetime.now()
    print(f"Time format: {timeformat} (current time as an example: {now.strftime(timeformat)})")
    return timeformat


def get_collection_names_from_vdirs(vdirs):
    names = []
    for name, path, vtype in sorted(vdirs or ()):
        if vtype == "discover":
            for vpath in utils.get_all_vdirs(utils.expand_path(path)):
                names.append(utils.get_unique_name(vpath, names))
        else:
            names.append(name)
    return names


def choose_default_calendar(vdirs):
    names = get_collection_names_from_vdirs(vdirs)

    print("Which calendar do you want as a default calendar?")
    print("(The default calendar is used when no calendar is specified.)")
    print(f"Configured calendars: {', '.join(names)}")
    default_calendar = prompt(
        "Please type one of the above options",
        default=names[0],
        type=Choice(names),
    )
    return default_calendar


def get_vdirs_from_vdirsyncer_config():
    """trying to load vdirsyncer's config and read all vdirs from it"""
    try:
        from vdirsyncer.cli import config
        from vdirsyncer.exceptions import UserError
    except ImportError:
        print("Couldn't load vdirsyncer to discover its calendars.")
        return None
    try:
        vdir_config = config.load_config()
    except UserError as error:
        print("Sorry, loading vdirsyncer config failed with the following error message:")
        print(error)
        return None
    vdirs = []
    for storage in vdir_config.storages.values():
        if storage["type"] == "filesystem":
            # TODO detect type of storage properly
            path = storage["path"]
            if path[-1] != "/":
                path += "/"
            path += "*"
            vdirs.append((storage["instance_name"], path, "discover"))
    if vdirs == []:
        print("No calendars found from vdirsyncer.")
        return None
    else:
        return vdirs


def find_vdir():
    """Use one or more existing vdirs on the system.

    Tries to get data from vdirsyncer if it's installed and configured, and
    asks user to confirm it. If not, prompt the user for the path to a single
    vdir.
    """
    print("The following calendars were found:")
    synced_vdirs = get_vdirs_from_vdirsyncer_config()
    if synced_vdirs:
        print(f"Found {len(synced_vdirs)} calendars from vdirsyncer")
        for name, path, _ in synced_vdirs:
            print(f"  {name}: {compressuser(path)}")
        if confirm("Use these calendars for khal?", default=True):
            return synced_vdirs

    vdir_path = prompt("Enter the path to a vdir calendar")
    vdir_path = normpath(expanduser(expandvars(vdir_path)))
    return [("private", vdir_path, "calendar")]


def create_vdir(names=None):
    """create a new vdir, make sure the name doesn't collide with existing names

    :param names: names of existing vdirs
    """
    names = names or []
    name = "private"
    while True:
        path = join(xdg.BaseDirectory.xdg_data_home, "khal", "calendars", name)
        path = normpath(expanduser(expandvars(path)))
        if name not in names and not exists(path):
            break
        else:
            name += "1"
    try:
        makedirs(path)
    except OSError as error:
        print(f"Could not create directory {path} because of {error}. Exiting")
        raise
    print(f"Created new vdir at {path}")
    return [(name, path, "calendar")]


# Parsing and then dumping config naively could lose comments and formatting.
# Since we don't need to modify existing fields, we can simply append our new
# config to the end of the file.
VDS_CONFIG_START = """\
[general]
status_path = "~/.local/share/vdirsyncer/status/"
"""

VDS_CONFIG_TEMPLATE = """
[pair khal_pair_{pairno}]
a = "khal_pair_{pairno}_local"
b = "khal_pair_{pairno}_remote"
collections = ["from a", "from b"]

[storage khal_pair_{pairno}_local]
type = "filesystem"
path = {local_path}
fileext = ".ics"

[storage khal_pair_{pairno}_remote]
type = "caldav"
url = {url}
username = {username}
password = {password}
"""


def vdirsyncer_config_path():
    """Find where vdirsyncer will look for it's config.

    There may or may not already be a file at the returned path.
    """
    fname = environ.get("VDIRSYNCER_CONFIG", None)
    if fname is None:
        fname = normpath(expanduser("~/.vdirsyncer/config"))
        if not exists(fname):
            xdg_config_dir = environ.get("XDG_CONFIG_HOME", normpath(expanduser("~/.config/")))
            fname = join(xdg_config_dir, "vdirsyncer/config")
    return fname


def get_available_pairno():
    """Find N so that 'khal_pair_N' is not already used in vdirsyncer config"""
    try:
        from vdirsyncer.cli import config
    except ImportError:
        raise FatalError("vdirsyncer config exists, but couldn't import vdirsyncer.")
    vdir_config = config.load_config()
    pairno = 1
    while f"khal_pair_{pairno}" in vdir_config.pairs:
        pairno += 1
    return pairno


def create_synced_vdir():
    """Create a new vdir, and set up vdirsyncer to sync it."""
    name, path, _ = create_vdir()[0]

    caldav_url = prompt("CalDAV URL")
    username = prompt("Username")
    password = prompt("Password", hide_input=True)

    vds_config = vdirsyncer_config_path()
    if exists(vds_config):
        # We are adding a pair to vdirsyncer config
        mode = "a"
        new_file = False
        pairno = get_available_pairno()
    else:
        # We're setting up vdirsyncer for the first time
        mode = "w"
        new_file = True
        pairno = 1

    with open(vds_config, mode) as f:
        if new_file:
            f.write(VDS_CONFIG_START)

        f.write(
            VDS_CONFIG_TEMPLATE.format(
                local_path=json.dumps(dirname(path)),
                url=json.dumps(caldav_url),
                username=json.dumps(username),
                password=json.dumps(password),
                pairno=pairno,
            )
        )
    start_syncing()
    return [(name, path, "calendar")]


def start_syncing():
    """Run vdirsyncer to sync the newly created vdir with the remote."""
    print("Syncing calendar...")
    try:
        exit_code = call(["vdirsyncer", "discover"])
    except FileNotFoundError:
        print("Could not find vdirsyncer - please set it up manually")
    else:
        if exit_code == 0:
            exit_code = call(["vdirsyncer", "sync"])
        if exit_code != 0:
            print("vdirsyncer failed - please set up sync manually")

    # Add code here to check platform and automatically set up cron or similar
    print(
        "Please set up your system to run 'vdirsyncer sync' periodically, "
        "using cron or similar mechanisms."
    )


def choose_vdir_calendar():
    """query the user for their preferred calendar source"""
    choices = [
        ("Create a new calendar on this computer", create_vdir),
        ("Use a calendar already on this computer (vdir format)", find_vdir),
        (
            "Sync a calendar from the internet (CalDAV format, requires vdirsyncer)",
            create_synced_vdir,
        ),
    ]
    validate = partial(validate_int, min_value=0, max_value=2)
    for i, (desc, _func) in enumerate(choices):
        print(f"[{i}] {desc}")
    choice_no = prompt("Please choose one of the above options", value_proc=validate)
    return choices[choice_no][1]()


def create_config(vdirs, dateformat, timeformat, default_calendar=None):
    config = ["[calendars]"]
    for name, path, type_ in sorted(vdirs or ()):
        config.append(f"\n[[{name}]]")
        config.append(f"path = {path}")
        config.append(f"type = {type_}")

    config.append("\n[locale]")
    config.append(
        f"timeformat = {timeformat}\n"
        f"dateformat = {dateformat}\n"
        f"longdateformat = {dateformat}\n"
        f"datetimeformat = {dateformat} {timeformat}\n"
        f"longdatetimeformat = {dateformat} {timeformat}\n"
    )
    if default_calendar:
        config.append("[default]")
        config.append(f"default_calendar = {default_calendar}\n")
    config = "\n".join(config)

    return config


def configwizard():
    config_file = find_configuration_file()
    if config_file is not None:
        logger.fatal(f"Found an existing config file at {compressuser(config_file)}.")
        logger.fatal(
            "If you want to create a new configuration file, "
            "please remove the old one first. Exiting."
        )
        raise FatalError()
    dateformat = choose_datetime_format()
    print()
    timeformat = choose_time_format()
    print()
    try:
        vdirs = choose_vdir_calendar()
    except OSError as error:
        raise FatalError(error)

    if not vdirs:
        print("\nWARNING: no vdir configured, khal will not be usable like this!\n")

    print()
    if vdirs:
        default_calendar = choose_default_calendar(vdirs)
    else:
        default_calendar = None

    config = create_config(
        vdirs,
        dateformat=dateformat,
        timeformat=timeformat,
        default_calendar=default_calendar,
    )
    config_path = join(xdg.BaseDirectory.xdg_config_home, "khal", "config")
    if not confirm(
        f"Do you want to write the config to {compressuser(config_path)}? "
        "(Choosing `No` will abort)",
        default=True,
    ):
        raise FatalError("User aborted...")
    config_dir = join(xdg.BaseDirectory.xdg_config_home, "khal")
    if not exists(config_dir) and not isdir(config_dir):
        try:
            makedirs(config_dir)
        except OSError as error:
            print(
                f"Could not write config file at {compressuser(config_dir)} because of "
                f"{error}. Aborting"
            )
            raise FatalError(error)
        else:
            print(f"created directory {compressuser(config_dir)}")
    with open(config_path, "w") as config_file:
        config_file.write(config)
    print(f"Successfully wrote configuration to {compressuser(config_path)}")



# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""collection of icalendar helper functions"""

import datetime as dt
import logging
from collections import defaultdict
from hashlib import sha256

import dateutil.rrule
import icalendar
import pytz

from .exceptions import UnsupportedRecurrence
from .parse_datetime import rrulefstr
from .utils import generate_random_uid, localize_strip_tz, str2alarm, to_unix_time

logger = logging.getLogger("khal")

# Force use of pytz because we rely on functionalities not available in
# zoneinfo.
icalendar.use_pytz()


def split_ics(ics: str, random_uid: bool = False, default_timezone=None) -> list:
    """split an ics string into several according to VEVENT's UIDs

    and sort the right VTIMEZONEs accordingly
    ignores all other ics components
    :param random_uid: assign random uids to all events
    """
    cal = cal_from_ics(ics)
    tzs = {}

    events_grouped = defaultdict(list)
    for item in cal.walk():
        # Since some events could have a Windows format timezone (e.g. 'New Zealand
        # Standard Time' for 'Pacific/Auckland' in Olson format), we convert any
        # Windows format timezones to Olson.
        if item.name == "VTIMEZONE":
            if item["TZID"] in icalendar.timezone.windows_to_olson.WINDOWS_TO_OLSON:
                key = icalendar.timezone.windows_to_olson.WINDOWS_TO_OLSON[item["TZID"]]
            else:
                key = item["TZID"]
            tzs[key] = item

        if item.name == "VEVENT":
            if "UID" not in item:
                logger.warning(
                    f"Event with summary '{item['SUMMARY']}' doesn't have a unique ID."
                    "A generated ID will be used instead."
                )
                item["UID"] = sha256(item.to_ical()).hexdigest()
            events_grouped[item["UID"]].append(item)
        else:
            continue
    out = []
    saved_exception = None
    for uid, events in sorted(events_grouped.items()):
        try:
            ics = ics_from_list(events, tzs, random_uid, default_timezone)
        except Exception as exception:
            logger.warning(f"Error when trying to import the event {uid}")
            saved_exception = exception
        else:
            out.append(ics)
    if saved_exception:
        raise saved_exception
    return out


def new_vevent(
    locale,
    dtstart: dt.date,
    dtend: dt.date,
    summary: str,
    timezone: pytz.BaseTzInfo | None = None,
    allday: bool = False,
    description: str | None = None,
    location: str | None = None,
    categories: list[str] | str | None = None,
    repeat: str | None = None,
    until=None,
    alarms: str | None = None,
    url: str | None = None,
) -> icalendar.Event:
    """create a new event

    :param dtstart: starttime of that event
    :param dtend: end time of that event, if this is a *date*, this value is
        interpreted as being the last date the event is scheduled on, i.e.
        the VEVENT DTEND will be *one day later*
    :param summary: description of the event, used in the SUMMARY property
    :param timezone: timezone of the event (start and end)
    :param allday: if set to True, we will not transform dtstart and dtend to
        datetime
    :param url: url of the event
    :returns: event
    """
    if not allday and timezone is not None:
        assert isinstance(dtstart, dt.datetime)
        assert isinstance(dtend, dt.datetime)
        dtstart = timezone.localize(dtstart)
        dtend = timezone.localize(dtend)

    event = icalendar.Event()
    event.add("dtstart", dtstart)
    event.add("dtend", dtend)
    event.add("dtstamp", dt.datetime.now())
    event.add("summary", summary)
    event.add("uid", generate_random_uid())
    # event.add('sequence', 0)

    if description:
        event.add("description", description)
    if location:
        event.add("location", location)
    if categories:
        event.add("categories", categories)
    if url:
        event.add("url", icalendar.vUri(url))
    if repeat and repeat != "none":
        rrule = rrulefstr(repeat, until, locale, getattr(dtstart, "tzinfo", None))
        event.add("rrule", rrule)
    if alarms:
        for alarm in str2alarm(alarms, description or ""):
            event.add_component(alarm)
    return event


def ics_from_list(
    events: list[icalendar.Event], tzs, random_uid: bool = False, default_timezone=None
) -> str:
    """convert an iterable of icalendar.Events to an icalendar str

    :params events: list of events all with the same uid
    :param random_uid: assign random uids to all events
    :param tzs: collection of timezones
    :type tzs: dict(icalendar.cal.Vtimzone
    """
    calendar = icalendar.Calendar()
    calendar.add("version", "2.0")
    calendar.add("prodid", "-//PIMUTILS.ORG//NONSGML khal / icalendar //EN")

    if random_uid:
        new_uid = generate_random_uid()

    needed_tz, missing_tz = set(), set()
    for sub_event in events:
        sub_event = sanitize(sub_event, default_timezone=default_timezone)
        if random_uid:
            sub_event["UID"] = new_uid
        # icalendar round-trip converts `TZID=a b` to `TZID="a b"` investigate, file bug XXX
        for prop in ["DTSTART", "DTEND", "DUE", "EXDATE", "RDATE", "RECURRENCE-ID", "DUE"]:
            if isinstance(sub_event.get(prop), list):
                items = sub_event.get(prop)
            else:
                items = [sub_event.get(prop)]

            for item in items:
                if not (hasattr(item, "dt") or hasattr(item, "dts")):
                    continue
                # if prop is a list, all items have the same parameters
                datetime_ = item.dts[0].dt if hasattr(item, "dts") else item.dt
                if not hasattr(datetime_, "tzinfo"):
                    continue
                # check for datetimes' timezones which are not understood by
                # icalendar
                if (
                    datetime_.tzinfo is None
                    and "TZID" in item.params
                    and item.params["TZID"] not in missing_tz
                ):
                    logger.warning(
                        f"Cannot find timezone `{item.params['TZID']}` in .ics file, "
                        "using default timezone. This can lead to erroneous time shifts"
                    )
                    missing_tz.add(item.params["TZID"])
                elif (
                    datetime_.tzinfo
                    and datetime_.tzinfo != pytz.UTC
                    and datetime_.tzinfo not in needed_tz
                ):
                    needed_tz.add(datetime_.tzinfo)

    for tzid in needed_tz:
        if str(tzid) in tzs:
            calendar.add_component(tzs[str(tzid)])
        else:
            logger.warning(
                f"Cannot find timezone `{tzid}` in .ics file, this could be a bug, "
                "please report this issue at http://github.com/pimutils/khal/."
            )
    for sub_event in events:
        calendar.add_component(sub_event)
    return calendar.to_ical().decode("utf-8")


def expand(
    vevent: icalendar.Event,
    href: str = "",
) -> list[tuple[dt.datetime, dt.datetime]] | None:
    """
    Constructs a list of start and end dates for all recurring instances of the
    event defined in vevent.

    It considers RRULE as well as RDATE and EXDATE properties. In case of
    unsupported recursion rules an UnsupportedRecurrence exception is thrown.

    If the vevent contains a RECURRENCE-ID property, no expansion is done,
    the function still returns a tuple of start and end (date)times.

    :param vevent: vevent to be expanded
    :param href: the href of the vevent, used for more informative logging and
                 nothing else
    :returns: list of start and end (date)times of the expanded event
    """
    # we do this now and than never care about the "real" end time again
    if "DURATION" in vevent:
        duration = vevent["DURATION"].dt
    else:
        duration = vevent["DTEND"].dt - vevent["DTSTART"].dt

    # if this vevent has a RECURRENCE_ID property, no expansion will be
    # performed
    expand = not bool(vevent.get("RECURRENCE-ID"))

    dtstart_prop = vevent["DTSTART"]
    # Check for VALUE=DATE parameter to detect all-day events
    # icalendar>=7.0 version returns datetime with tzinfo even for DATE values
    allday = dtstart_prop.params.get("VALUE") == "DATE" or not isinstance(
        dtstart_prop.dt, dt.datetime
    )
    # Don't use timezone for all-day events
    events_tz = getattr(dtstart_prop.dt, "tzinfo", None) if not allday else None

    def sanitize_datetime(date: dt.date) -> dt.date:
        if allday and isinstance(date, dt.datetime):
            date = date.date()
        if events_tz is not None:
            date = events_tz.localize(date)
        return date

    rrule_param = vevent.get("RRULE")
    if expand and rrule_param is not None:
        vevent = sanitize_rrule(vevent)

        # dst causes problem while expanding the rrule, therefore we transform
        # everything to naive datetime objects and transform back after
        # expanding
        # See https://github.com/dateutil/dateutil/issues/102
        dtstart = vevent["DTSTART"].dt
        if events_tz:
            dtstart = dtstart.replace(tzinfo=None)
        # For all-day events, ensure dtstart is a date, not datetime.
        if allday and isinstance(dtstart, dt.datetime):
            dtstart = dtstart.date()

        rrule = dateutil.rrule.rrulestr(
            rrule_param.to_ical().decode(),
            dtstart=dtstart,
            ignoretz=True,
        )

        # telling mypy, that _until exists
        # we are very sure (TM) that rrulestr always returns a rrule, not a
        # rruleset (which wouldn't have a _until attribute)
        if rrule._until is None:  # type: ignore
            # rrule really doesn't like to calculate all recurrences until
            # eternity, so we only do it until 2037, because a) I'm not sure
            # if python can deal with larger datetime values yet and b) pytz
            # doesn't know any larger transition times
            rrule._until = dt.datetime(2037, 12, 31)  # type: ignore
        else:
            if events_tz and "Z" in rrule_param.to_ical().decode():
                assert isinstance(rrule._until, dt.datetime)  # type: ignore
                rrule._until = (  # type: ignore
                    pytz.UTC.localize(rrule._until).astimezone(events_tz).replace(tzinfo=None)  # type: ignore
                )

            # rrule._until and dtstart could be dt.date or dt.datetime. They
            # need to be the same for comparison
            testuntil = rrule._until  # type: ignore
            if type(dtstart) is dt.date and type(testuntil) is dt.datetime:
                testuntil = testuntil.date()
            teststart = dtstart
            if type(testuntil) is dt.date and type(teststart) is dt.datetime:
                teststart = teststart.date()

            if testuntil < teststart:
                logger.warning(
                    f"{href}: Unsupported recurrence. UNTIL is before DTSTART.\n"
                    "This event will not be available in khal."
                )
                return None

        if rrule.count() == 0:
            logger.warning(
                f"{href}: Recurrence defined but will never occur.\n"
                "This event will not be available in khal."
            )
            return None

        rrule = map(sanitize_datetime, rrule)  # type: ignore

        logger.debug(f"calculating recurrence dates for {href}, this might take some time.")

        # RRULE and RDATE may specify the same date twice, it is recommended by
        # the RFC to consider this as only one instance
        dtstartl = set(rrule)
        if not dtstartl:
            raise UnsupportedRecurrence()
    else:
        dtstartl = {vevent["DTSTART"].dt}

    def get_dates(vevent, key):
        # TODO replace with get_all_properties
        dates = vevent.get(key)
        if dates is None:
            return
        if not isinstance(dates, list):
            dates = [dates]

        dates = (leaf.dt for tree in dates for leaf in tree.dts)
        dates = localize_strip_tz(dates, events_tz)
        return map(sanitize_datetime, dates)

    # include explicitly specified recursion dates
    if expand:
        dtstartl.update(get_dates(vevent, "RDATE") or ())

    # remove excluded dates
    if expand:
        for date in get_dates(vevent, "EXDATE") or ():
            try:
                dtstartl.remove(date)
            except KeyError:
                logger.warning(
                    f"In event {href}, excluded instance starting at {date} "
                    "not found, event might be invalid."
                )

    dtstartend = [(start, start + duration) for start in dtstartl]
    # not necessary, but I prefer deterministic output
    dtstartend.sort()
    return dtstartend


def assert_only_one_uid(cal: icalendar.Calendar):
    """assert that all VEVENTs in cal have the same UID"""
    uids = set()
    for item in cal.walk():
        if item.name == "VEVENT":
            uids.add(item["UID"])
    if len(uids) > 1:
        return False
    else:
        return True


def sanitize(
    vevent: icalendar.Event,
    default_timezone: pytz.BaseTzInfo,
    href: str = "",
    calendar: str = "",
) -> icalendar.Event:
    """
    clean up vevents we do not understand

    :param vevent: the vevent that needs to be cleaned
    :param default_timezone: timezone to apply to start and/or end dates which
         were supposed to be localized but which timezone was not understood
         by icalendar
    :param href: used for logging to inform user which .ics files are
        problematic
    :param calendar: used for logging to inform user which .ics files are
        problematic
    :returns: clean vevent
    """
    # convert localized datetimes with timezone information we don't
    # understand to the default timezone
    # TODO do this for everything where a TZID can appear (RDATE, EXDATE)
    for prop in ["DTSTART", "DTEND", "DUE", "RECURRENCE-ID"]:
        if prop in vevent and invalid_timezone(vevent[prop]):
            timezone = vevent[prop].params.get("TZID")
            value = default_timezone.localize(vevent.pop(prop).dt)
            vevent.add(prop, value)
            logger.warning(
                f"{prop} localized in invalid or incomprehensible timezone "
                f"`{timezone}` in {calendar}/{href}. This could lead to this "
                "event being wrongly displayed."
            )

    vdtstart = vevent.pop("DTSTART", None)
    vdtend = vevent.pop("DTEND", None)
    dtstart = getattr(vdtstart, "dt", None)
    dtend = getattr(vdtend, "dt", None)

    # event with missing DTSTART
    if dtstart is None:
        raise ValueError("Event has no start time (DTSTART).")
    dtstart, dtend = sanitize_timerange(dtstart, dtend, duration=vevent.get("DURATION", None))

    vevent.add("DTSTART", dtstart)
    if dtend is not None:
        vevent.add("DTEND", dtend)
    return vevent


def sanitize_timerange(dtstart, dtend, duration=None):
    """return sensible dtstart and end for events that have an invalid or
    missing DTEND, assuming the event just lasts one hour."""

    if isinstance(dtstart, dt.datetime) and isinstance(dtend, dt.datetime):
        if dtstart.tzinfo and not dtend.tzinfo:
            logger.warning(
                "Event end time has no timezone. Assuming it's the same timezone as the start time"
            )
            dtend = dtstart.tzinfo.localize(dtend)
        if not dtstart.tzinfo and dtend.tzinfo:
            logger.warning(
                "Event start time has no timezone. Assuming it's the same timezone as the end time"
            )
            dtstart = dtend.tzinfo.localize(dtstart)
    if dtend is not None and type(dtstart) is not type(dtend):
        raise ValueError(
            "The event's end time (DTEND) and start time (DTSTART) are not of the same type."
        )

    if dtend is None and duration is None:
        if isinstance(dtstart, dt.datetime):
            dtend = dtstart + dt.timedelta(hours=1)
        else:
            dtend = dtstart + dt.timedelta(days=1)
    elif dtend is not None:
        if dtend < dtstart:
            raise ValueError(
                "The event's end time (DTEND) is older than the event's start time (DTSTART)."
            )
        elif dtend == dtstart:
            logger.warning(
                "Event start time and end time are the same. "
                "Assuming the event's duration is one hour."
            )
            if isinstance(dtstart, dt.datetime):
                dtend += dt.timedelta(hours=1)
            else:
                dtend += dt.timedelta(days=1)

    return dtstart, dtend


def sanitize_rrule(vevent):
    """fix problems with RRULE:UNTIL"""
    if "rrule" in vevent and "UNTIL" in vevent["rrule"]:
        until = vevent["rrule"]["UNTIL"][0]
        dtstart = vevent["dtstart"].dt
        # DTSTART is date, UNTIL is datetime
        if not isinstance(dtstart, dt.datetime) and isinstance(until, dt.datetime):
            vevent["rrule"]["until"] = until.date()
    return vevent


def invalid_timezone(prop):
    """check if an icalendar property has a timezone attached we don't understand"""
    return bool(hasattr(prop.dt, "tzinfo") and prop.dt.tzinfo is None and "TZID" in prop.params)


def _get_all_properties(vevent, prop):
    """Get all properties from a vevent, even if there are several entries

    example input:
    EXDATE:1234,4567
    EXDATE:7890

    returns: [1234, 4567, 7890]

    :type vevent: icalendar.cal.Event
    :type prop: str
    """
    if prop not in vevent:
        return []
    if isinstance(vevent[prop], list):
        rdates = [leaf.dt for tree in vevent[prop] for leaf in tree.dts]
    else:
        rdates = [vddd.dt for vddd in vevent[prop].dts]
    return rdates


def delete_instance(vevent: icalendar.Event, instance: dt.datetime) -> None:
    """remove a recurrence instance from a VEVENT's RRDATE list or add it
    to the EXDATE list
    """
    # TODO check where this instance is coming from and only call the
    # appropriate function
    if "RRULE" in vevent:
        exdates = _get_all_properties(vevent, "EXDATE")
        exdates += [instance]
        vevent.pop("EXDATE")
        vevent.add("EXDATE", exdates)
    if "RDATE" in vevent:
        rdates = [one for one in _get_all_properties(vevent, "RDATE") if one != instance]
        vevent.pop("RDATE")
        if rdates != []:
            vevent.add("RDATE", rdates)


def sort_key(vevent: icalendar.Event) -> tuple[str, float]:
    """helper function to determine order of VEVENTS
    so that recurrence-id events come after the corresponding rrule event, etc
    :param vevent: icalendar.Event
    :rtype: tuple(str, int)
    """
    assert isinstance(vevent, icalendar.Event)
    uid = str(vevent["UID"])
    rec_id = vevent.get("RECURRENCE-ID")
    if rec_id is None:
        return uid, 0
    rrange = rec_id.params.get("RANGE")
    if rrange == "THISANDFUTURE":
        return uid, to_unix_time(rec_id.dt)
    else:
        return uid, 1


def cal_from_ics(ics: str) -> icalendar.Calendar:
    """
    :param ics: an icalendar formatted string
    """
    try:
        cal = icalendar.Calendar.from_ical(ics)
    except ValueError as error:
        if (
            len(error.args) > 0
            and isinstance(error.args[0], str)
            and error.args[0].startswith("Offset must be less than 24 hours")
        ):
            logger.warning(
                "Invalid timezone offset encountered, "
                "timezone information may be wrong: " + str(error.args[0])
            )
            icalendar.vUTCOffset.ignore_exceptions = True
            cal = icalendar.Calendar.from_ical(ics)
            icalendar.vUTCOffset.ignore_exceptions = False
        else:
            raise
    return cal



# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import calendar
import datetime as dt
from locale import LC_ALL, LC_TIME, getlocale, setlocale

from click import style

from .khalendar import CalendarCollection
from .terminal import colored
from .utils import get_month_abbr_len

setlocale(LC_ALL, "")


def get_weekheader(firstweekday: int) -> str:
    try:
        mylocale = ".".join(getlocale(LC_TIME))  # type: ignore
    except TypeError:
        mylocale = "C"

    _calendar = calendar.LocaleTextCalendar(firstweekday, locale=mylocale)  # type: ignore
    return _calendar.formatweekheader(2)


def getweeknumber(date: dt.date) -> int:
    """return iso week number for datetime.date object
    :param date: date
    :return: weeknumber
    """
    return dt.date.isocalendar(date)[1]


def get_calendar_color(calendar: str, default_color: str, collection: CalendarCollection) -> str:
    """Because multi-line lambdas would be un-Pythonic"""
    if collection._calendars[calendar]["color"] == "":
        return default_color
    return collection._calendars[calendar]["color"]


def get_color_list(
    calendars: list[str], default_color: str, collection: CalendarCollection
) -> list[str]:
    """Get the list of possible colors for the day, taking into account priority"""
    dcolors = [
        (
            get_calendar_color(x, default_color, collection),
            collection._calendars[x]["priority"],
        )
        for x in calendars
    ]

    dcolors.sort(key=lambda x: x[1], reverse=True)

    maxPriority = dcolors[0][1]
    return list({x[0] for x in filter(lambda x: x[1] == maxPriority, dcolors)})


def str_highlight_day(
    day: dt.date,
    calendars: list[str],
    hmethod: str | None,
    default_color: str,
    multiple: str,
    multiple_on_overflow: bool,
    color: str,
    bold_for_light_color: bool,
    collection: CalendarCollection,
) -> str:
    """returns a string with day highlighted according to configuration"""
    dstr = str(day.day).rjust(2)
    if color == "":
        dcolors = get_color_list(calendars, default_color, collection)
        if len(dcolors) > 1:
            if multiple == "" or (multiple_on_overflow and len(dcolors) == 2):
                if hmethod == "foreground" or hmethod == "fg":
                    return colored(
                        dstr[:1], fg=dcolors[0], bold_for_light_color=bold_for_light_color
                    ) + colored(dstr[1:], fg=dcolors[1], bold_for_light_color=bold_for_light_color)
                else:
                    return colored(
                        dstr[:1], bg=dcolors[0], bold_for_light_color=bold_for_light_color
                    ) + colored(dstr[1:], bg=dcolors[1], bold_for_light_color=bold_for_light_color)
            else:
                dcolor = multiple
        else:
            dcolor = dcolors[0] or default_color
    else:
        dcolor = color
    if dcolor != "":
        if hmethod == "foreground" or hmethod == "fg":
            return colored(dstr, fg=dcolor, bold_for_light_color=bold_for_light_color)
        else:
            return colored(dstr, bg=dcolor, bold_for_light_color=bold_for_light_color)
    return dstr


def str_week(
    week: list[dt.date],
    today: dt.date,
    collection: CalendarCollection | None = None,
    hmethod: str | None = None,
    default_color: str = "",
    multiple: str = "",
    multiple_on_overflow: bool = False,
    color: str = "",
    highlight_event_days: bool = False,
    locale=None,
    bold_for_light_color: bool = True,
) -> str:
    """returns a string representing one week,
    if for day == today color is reversed

    :param week: list of 7 datetime.date objects (one week)
    :param today: the date of today
    :return: string, which if printed on terminal appears to have length 20,
             but may contain ascii escape sequences
    """
    strweek = ""
    if highlight_event_days and collection is None:
        raise ValueError(
            "if `highlight_event_days` is True, `collection` must be a CalendarCollection"
        )
    for day in week:
        if day == today:
            day_str = style(str(day.day).rjust(2), reverse=True)
        elif highlight_event_days:
            assert collection is not None
            devents = list(collection.get_calendars_on(day))
            if len(devents) > 0:
                day_str = str_highlight_day(
                    day,
                    devents,
                    hmethod,
                    default_color,
                    multiple,
                    multiple_on_overflow,
                    color,
                    bold_for_light_color,
                    collection,
                )
            else:
                day_str = str(day.day).rjust(2)
        else:
            day_str = str(day.day).rjust(2)
        strweek = strweek + day_str + " "
    return strweek


def vertical_month(
    month: int | None = None,
    year: int | None = None,
    today: dt.date | None = None,
    weeknumber: bool | str = False,
    count: int = 3,
    firstweekday: int = 0,
    monthdisplay: str = "firstday",
    collection=None,
    hmethod: str = "fg",
    default_color: str = "",
    multiple: str = "",
    multiple_on_overflow: bool = False,
    color: str = "",
    highlight_event_days: bool = False,
    locale=None,
    bold_for_light_color: bool = True,
) -> list[str]:
    """
    returns a list() of str() of weeks for a vertical arranged calendar

    :param month: first month of the calendar,
                  if non given, current month is assumed
    :param year: year of the first month included,
                 if non given, current year is assumed
    :param today: day highlighted, if non is given, current date is assumed
    :param weeknumber: if not False the iso weeknumber will be shown for each
                       week, if weeknumber is 'right' it will be shown in its
                       own column, if it is 'left' it will be shown interleaved
                       with the month names
    :returns: calendar strings,  may also include some
              ANSI (color) escape strings
    """
    if month is None:
        month = dt.date.today().month
    if year is None:
        year = dt.date.today().year
    if today is None:
        today = dt.date.today()

    khal = []
    w_number = "  " if weeknumber == "right" else ""
    calendar.setfirstweekday(firstweekday)
    weekheaders = get_weekheader(firstweekday)
    month_abbr_len = get_month_abbr_len()
    khal.append(style(" " * month_abbr_len + weekheaders + " " + w_number, bold=True))
    _calendar = calendar.Calendar(firstweekday)
    for _ in range(count):
        for week in _calendar.monthdatescalendar(year, month):
            if monthdisplay == "firstday":
                new_month = len([day for day in week if day.day == 1])
            else:
                new_month = len(week if week[0].day <= 7 else [])
            strweek = str_week(
                week,
                today,
                collection,
                hmethod,
                default_color,
                multiple,
                multiple_on_overflow,
                color,
                highlight_event_days,
                locale,
                bold_for_light_color,
            )
            if new_month:
                m_name = style(calendar.month_abbr[week[6].month].ljust(month_abbr_len), bold=True)
            elif weeknumber == "left":
                m_name = style(str(getweeknumber(week[0])).center(month_abbr_len), bold=True)
            else:
                m_name = " " * month_abbr_len
            if weeknumber == "right":
                w_number = style(f"{getweeknumber(week[0]):2}", bold=True)
            else:
                w_number = ""

            sweek = m_name + strweek + w_number
            if sweek != khal[-1]:
                khal.append(sweek)
        month = month + 1
        if month > 12:
            month = 1
            year = year + 1
    return khal



# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import datetime as dt
from collections.abc import Callable
from typing import TYPE_CHECKING, Literal

import urwid

from khal.utils import get_weekday_occurrence, get_wrapped_text

from .calendarwidget import CalendarWidget
from .widgets import (
    AlarmsEditor,
    CAttrMap,
    Choice,
    CPadding,
    DateConversionError,
    DateWidget,
    ExtendedEdit,
    NColumns,
    NListBox,
    NPile,
    PositiveIntEdit,
    TimeWidget,
    ValidatedEdit,
    button,
)

if TYPE_CHECKING:
    import khal.khalendar.event


class StartEnd:
    def __init__(self, startdate, starttime, enddate, endtime) -> None:
        """collecting some common properties"""
        self.startdate = startdate
        self.starttime = starttime
        self.enddate = enddate
        self.endtime = endtime


class CalendarPopUp(urwid.PopUpLauncher):
    def __init__(
        self,
        widget,
        on_date_change,
        weeknumbers: Literal["left", "right", False] = False,
        firstweekday=0,
        monthdisplay="firstday",
        keybindings=None,
    ) -> None:
        self._on_date_change = on_date_change
        self._weeknumbers = weeknumbers
        self._monthdisplay = monthdisplay
        self._firstweekday = firstweekday
        self._keybindings = {} if keybindings is None else keybindings
        super().__init__(widget)

    def keypress(self, size, key):
        if key == "enter":
            self.open_pop_up()
        else:
            return super().keypress(size, key)

    def create_pop_up(self):
        def on_change(new_date):
            self.base_widget.set_value(new_date)
            self._on_date_change(new_date)

        on_press = {
            "enter": lambda _, __: self.close_pop_up(),
            "esc": lambda _, __: self.close_pop_up(),
        }
        try:
            initial_date = self.base_widget._get_current_value()
        except DateConversionError:
            return None
        else:
            pop_up = CalendarWidget(
                on_change,
                self._keybindings,
                on_press,
                firstweekday=self._firstweekday,
                weeknumbers=self._weeknumbers,
                monthdisplay=self._monthdisplay,
                initial=initial_date,
            )
            pop_up = CAttrMap(pop_up, "calendar", " calendar focus")
            pop_up = CAttrMap(urwid.LineBox(pop_up), "calendar", "calendar focus")
            return pop_up

    def get_pop_up_parameters(self):
        width = 31 if self._weeknumbers == "right" else 28
        return {"left": 0, "top": 1, "overlay_width": width, "overlay_height": 8}


class DateEdit(urwid.WidgetWrap):
    """Widget that allows editing a Date.

    Will open a calendar when `enter` is pressed, pressing `enter` again will
    select that date.
    """

    def __init__(
        self,
        startdt: dt.date,
        dateformat: str = "%Y-%m-%d",
        on_date_change: Callable = lambda _: None,
        weeknumbers: Literal["left", "right", False] = False,
        firstweekday: int = 0,
        monthdisplay: Literal["firstday", "firstfullweek"] = "firstday",
        keybindings: dict[str, list[str]] | None = None,
    ) -> None:
        datewidth = len(startdt.strftime(dateformat))
        self._dateformat = dateformat
        if startdt is None:
            startdt = dt.date.today()
        self._edit = ValidatedEdit(
            dateformat=dateformat,
            EditWidget=DateWidget,
            validate=self._validate,
            edit_text=startdt.strftime(dateformat),
            on_date_change=on_date_change,
        )
        wrapped = CalendarPopUp(
            self._edit, on_date_change, weeknumbers, firstweekday, monthdisplay, keybindings
        )
        padded = CAttrMap(
            urwid.Padding(wrapped, align="left", width=datewidth, left=0, right=1),
            "calendar",
            "calendar focus",
        )
        super().__init__(padded)

    def _validate(self, text: str):
        try:
            _date = dt.datetime.strptime(text, self._dateformat).date()
        except ValueError:
            return False
        else:
            return _date

    @property
    def date(self):
        """Get currently entered date, or False, if input is invalid.

        :returns: the currently entered date
        :rtype: datetime.date
        """
        return self._validate(self._edit.get_edit_text())

    @date.setter
    def date(self, date):
        """Update text of this Widget.

        :type date: datetime.date
        """
        self._edit.set_edit_text(dt.date.strftime(self._dateformat))


class StartEndEditor(urwid.WidgetWrap):
    """Widget for editing start and end times (of an event)."""

    def __init__(
        self,
        start: dt.datetime,
        end: dt.datetime,
        conf,
        on_start_date_change=lambda x: None,
        on_end_date_change=lambda x: None,
        on_type_change: Callable[[bool], None] = lambda _: None,
    ) -> None:
        """
        :param on_start_date_change: a callable that gets called everytime a new
            start date is entered, with that new date as an argument
        :param on_end_date_change: same as for on_start_date_change, just for the
            end date
        :param on_type_change: callback that gets called when the event type
            (allday or datetime) changes, gets passed True if toggled to allday
            and False if toggled to daytime
        """
        self.allday = not isinstance(start, dt.datetime)
        self.conf = conf
        self._startdt: dt.date = start
        self._original_start: dt.date = start
        self._enddt: dt.date = end
        self._original_end: dt.date = end
        self.on_start_date_change = on_start_date_change
        self.on_end_date_change = on_end_date_change
        self.on_type_change = on_type_change
        self._datewidth = len(start.strftime(self.conf["locale"]["longdateformat"]))
        self._timewidth = len(start.strftime(self.conf["locale"]["timeformat"]))
        # this will contain the widgets for [start|end] [date|time]
        self.widgets = StartEnd(None, None, None, None)
        self.checkallday = urwid.CheckBox("Allday", state=self.allday, on_state_change=self.toggle)
        self.toggle(None, self.allday)

    def keypress(self, size, key):
        return super().keypress(size, key)

    @property
    def startdt(self):
        if self.allday and isinstance(self._startdt, dt.datetime):
            return self._startdt.date()
        else:
            return self._startdt

    @property
    def _start_time(self):
        try:
            return self._startdt.time()
        except AttributeError:
            return dt.time(0)

    @property
    def localize_start(self):
        if getattr(self.startdt, "tzinfo", None) is None:
            return self.conf["locale"]["default_timezone"].localize
        else:
            return self.startdt.tzinfo.localize

    @property
    def localize_end(self):
        if getattr(self.enddt, "tzinfo", None) is None:
            return self.conf["locale"]["default_timezone"].localize
        else:
            return self.enddt.tzinfo.localize

    @property
    def enddt(self):
        if self.allday and isinstance(self._enddt, dt.datetime):
            return self._enddt.date()
        else:
            return self._enddt

    @property
    def _end_time(self):
        try:
            return self._enddt.time()
        except AttributeError:
            return dt.time(0)

    def _validate_start_time(self, text):
        try:
            startval = dt.datetime.strptime(text, self.conf["locale"]["timeformat"])
            self._startdt = self.localize_start(
                dt.datetime.combine(self._startdt.date(), startval.time())
            )
        except ValueError:
            return False
        else:
            return startval

    def _start_date_change(self, date):
        self._startdt = self.localize_start(dt.datetime.combine(date, self._start_time))
        self.on_start_date_change(date)

    def _validate_end_time(self, text):
        try:
            endval = dt.datetime.strptime(text, self.conf["locale"]["timeformat"])
            self._enddt = self.localize_end(dt.datetime.combine(self._enddt.date(), endval.time()))
        except ValueError:
            return False
        else:
            return endval

    def _end_date_change(self, date):
        self._enddt = self.localize_end(dt.datetime.combine(date, self._end_time))
        self.on_end_date_change(date)

    def toggle(self, checkbox, state: bool):
        """change from allday to datetime event

        :param checkbox: the checkbox instance that is used for toggling, gets
                         automatically passed by urwid (is not used)
        :type checkbox: checkbox
        :param state: state the event will toggle to;
                      True if allday event, False if datetime
        """

        if self.allday is True and state is False:
            self._startdt = dt.datetime.combine(self._startdt, dt.datetime.min.time())
            self._enddt = dt.datetime.combine(self._enddt, dt.datetime.min.time())
        elif self.allday is False and state is True:
            assert isinstance(self._startdt, dt.datetime)
            assert isinstance(self._enddt, dt.datetime)
            self._startdt = self._startdt.date()
            self._enddt = self._enddt.date()
        self.allday = state
        self.widgets.startdate = DateEdit(
            self._startdt,
            self.conf["locale"]["longdateformat"],
            self._start_date_change,
            self.conf["locale"]["weeknumbers"],
            self.conf["locale"]["firstweekday"],
            self.conf["view"]["monthdisplay"],
            self.conf["keybindings"],
        )
        self.widgets.enddate = DateEdit(
            self._enddt,
            self.conf["locale"]["longdateformat"],
            self._end_date_change,
            self.conf["locale"]["weeknumbers"],
            self.conf["locale"]["firstweekday"],
            self.conf["view"]["monthdisplay"],
            self.conf["keybindings"],
        )

        if state is True:  # allday event
            self.on_type_change(True)
            timewidth = 1
            self.widgets.starttime = urwid.Text("")
            self.widgets.endtime = urwid.Text("")
        elif state is False:  # datetime event
            self.on_type_change(False)
            timewidth = self._timewidth + 1
            raw_start_time_widget = ValidatedEdit(
                dateformat=self.conf["locale"]["timeformat"],
                EditWidget=TimeWidget,
                validate=self._validate_start_time,
                edit_text=self.startdt.strftime(self.conf["locale"]["timeformat"]),
            )
            self.widgets.starttime = urwid.Padding(
                raw_start_time_widget, align="left", width=self._timewidth + 1, left=1
            )

            raw_end_time_widget = ValidatedEdit(
                dateformat=self.conf["locale"]["timeformat"],
                EditWidget=TimeWidget,
                validate=self._validate_end_time,
                edit_text=self.enddt.strftime(self.conf["locale"]["timeformat"]),
            )
            self.widgets.endtime = urwid.Padding(
                raw_end_time_widget, align="left", width=self._timewidth + 1, left=1
            )

        columns = NPile(
            [
                self.checkallday,
                NColumns(
                    [
                        (5, urwid.Text("From:")),
                        (self._datewidth, self.widgets.startdate),
                        (timewidth, self.widgets.starttime),
                    ],
                    dividechars=1,
                ),
                NColumns(
                    [
                        (5, urwid.Text("To:")),
                        (self._datewidth, self.widgets.enddate),
                        (timewidth, self.widgets.endtime),
                    ],
                    dividechars=1,
                ),
            ],
            focus_item=1,
        )
        urwid.WidgetWrap.__init__(self, columns)

    @property
    def changed(self):
        """returns True if content has been edited, False otherwise"""
        return (self.startdt != self._original_start) or (self.enddt != self._original_end)

    def validate(self):
        return self.startdt <= self.enddt


class EventEditor(urwid.WidgetWrap):
    """Widget that allows Editing one `Event()`"""

    def __init__(
        self,
        pane,
        event: "khal.khalendar.event.Event",
        save_callback=None,
        always_save: bool = False,
    ) -> None:
        """
        :param save_callback: call when saving event with new start and end
             dates and recursiveness of original and edited event as parameters
        :type save_callback: callable
        :param always_save: save event even if it has not changed
        """
        self.pane = pane
        self.event = event
        self._save_callback = save_callback

        self.collection = pane.collection
        self._conf = pane._conf

        self._abort_confirmed = False

        self.description = event.description
        self.location = event.location
        self.attendees = event.attendees
        self.categories = event.categories
        self.url = event.url
        self.startendeditor = StartEndEditor(
            event.start_local,
            event.end_local,
            self._conf,
            self.start_datechange,
            self.end_datechange,
            self.type_change,
        )
        # TODO make sure recurrence rules cannot be edited if we only
        # edit one instance (or this and future) (once we support that)
        self.recurrenceeditor = RecurrenceEditor(
            self.event.recurobject,
            self._conf,
            event.start_local,
        )
        self.summary = urwid.AttrMap(
            ExtendedEdit(caption=("caption", "Title:       "), edit_text=event.summary),
            "edit",
            "edit focus",
        )

        divider = urwid.Divider(" ")

        def decorate_choice(c) -> tuple[str, str]:
            return ("calendar " + c["name"] + " popup", c["name"])

        self.calendar_chooser = CAttrMap(
            Choice(
                [self.collection._calendars[c] for c in self.collection.writable_names],
                self.collection._calendars[self.event.calendar],
                decorate_choice,
            ),
            "caption",
        )

        self.description = urwid.AttrMap(
            ExtendedEdit(
                caption=("caption", "Description: "), edit_text=self.description, multiline=True
            ),
            "edit",
            "edit focus",
        )
        self.location = urwid.AttrMap(
            ExtendedEdit(caption=("caption", "Location:    "), edit_text=self.location),
            "edit",
            "edit focus",
        )
        self.categories = urwid.AttrMap(
            ExtendedEdit(caption=("caption", "Categories:  "), edit_text=self.categories),
            "edit",
            "edit focus",
        )
        self.attendees = urwid.AttrMap(
            ExtendedEdit(
                caption=("caption", "Attendees:   "), edit_text=self.attendees, multiline=True
            ),
            "edit",
            "edit focus",
        )
        self.url = urwid.AttrMap(
            ExtendedEdit(caption=("caption", "URL:         "), edit_text=self.url),
            "edit",
            "edit focus",
        )
        self.alarmseditor: AlarmsEditor = AlarmsEditor(self.event)
        self.pile = NListBox(
            urwid.SimpleFocusListWalker(
                [
                    self.summary,
                    urwid.Columns(
                        [
                            (13, urwid.AttrMap(urwid.Text("Calendar:"), "caption")),
                            (12, self.calendar_chooser),
                        ],
                    ),
                    divider,
                    self.location,
                    self.categories,
                    self.description,
                    self.url,
                    divider,
                    self.attendees,
                    divider,
                    self.startendeditor,
                    self.recurrenceeditor,
                    divider,
                    self.alarmseditor,
                    divider,
                    urwid.Columns(
                        [(12, button("Save", on_press=self.save, padding_left=0, padding_right=0))]
                    ),
                    urwid.Columns(
                        [
                            (
                                12,
                                button(
                                    "Export", on_press=self.export, padding_left=0, padding_right=0
                                ),
                            )
                        ],
                    ),
                ]
            ),
            outermost=True,
        )
        self._always_save = always_save
        urwid.WidgetWrap.__init__(self, self.pile)

    def start_datechange(self, date):
        self.pane.eventscolumn.original_widget.set_focus_date(date)
        self.recurrenceeditor.update_startdt(date)

    def end_datechange(self, date):
        self.pane.eventscolumn.original_widget.set_focus_date(date)

    def type_change(self, allday: bool) -> None:
        """when the event type changes, we might want to change the default alarms

        :params allday: True if the event is now an allday event, False if it isn't
        """
        # test if self.alarmseditor exists
        if not hasattr(self, "alarmseditor"):
            return

        # to make the alarms before the event, we need to set it them to
        # negative values
        default_event_alarm = -1 * self._conf["default"]["default_event_alarm"]
        default_dayevent_alarm = -1 * self._conf["default"]["default_dayevent_alarm"]
        alarms = self.alarmseditor.get_alarms()
        if len(alarms) == 1:
            timedelta = alarms[0][0]
            if allday and timedelta == default_event_alarm:
                self.alarmseditor.clear()
                self.alarmseditor.add_alarm(None, default_dayevent_alarm)
            elif (not allday) and timedelta == default_dayevent_alarm:
                self.alarmseditor.clear()
                self.alarmseditor.add_alarm(None, default_event_alarm)
            else:
                # either there were more than one alarm or the alarm was not the default
                pass

    @property
    def title(self):  # Window title
        return f"Edit: {get_wrapped_text(self.summary)}"

    @classmethod
    def selectable(cls):
        return True

    @property
    def changed(self):
        if get_wrapped_text(self.summary) != self.event.summary:
            return True
        if get_wrapped_text(self.description) != self.event.description:
            return True
        if get_wrapped_text(self.location) != self.event.location:
            return True
        if get_wrapped_text(self.categories) != self.event.categories:
            return True
        if get_wrapped_text(self.url) != self.event.url:
            return True
        if get_wrapped_text(self.attendees) != self.event.attendees:
            return True
        if self.startendeditor.changed or self.calendar_chooser.changed:
            return True
        if self.recurrenceeditor.changed:
            return True
        if self.alarmseditor.changed:
            return True
        return False

    def update_vevent(self):
        self.event.update_summary(get_wrapped_text(self.summary))
        self.event.update_description(get_wrapped_text(self.description))
        self.event.update_location(get_wrapped_text(self.location))
        self.event.update_attendees(get_wrapped_text(self.attendees).split(","))
        self.event.update_categories(get_wrapped_text(self.categories).split(","))
        self.event.update_url(get_wrapped_text(self.url))

        if self.startendeditor.changed:
            self.event.update_start_end(self.startendeditor.startdt, self.startendeditor.enddt)
        if self.recurrenceeditor.changed:
            rrule = self.recurrenceeditor.active
            self.event.update_rrule(rrule)

        if self.alarmseditor.changed:
            self.event.update_alarms(self.alarmseditor.get_alarms())

    def export(self, button):
        """
        export the event as ICS
        :param button: not needed, passed via the button press
        """

        def export_this(_, user_data):
            try:
                self.event.export_ics(user_data.get_edit_text())
            except Exception as e:
                self.pane.window.backtrack()
                self.pane.window.alert(("light red", f"Failed to save event: {e}"))
                return

            self.pane.window.backtrack()
            self.pane.window.alert(("light green", "Event successfuly exported"))

        overlay = urwid.Overlay(
            ExportDialog(
                export_this,
                self.pane.window.backtrack,
                self.event,
            ),
            self.pane,
            "center",
            ("relative", 50),
            ("relative", 50),
            None,
        )
        self.pane.window.open(overlay)

    def save(self, button):
        """saves the event to the db

        (only when it has been changed or always_save is set)
        :param button: not needed, passed via the button press
        """
        if not self.startendeditor.validate():
            self.pane.window.alert(("light red", "Can't save: end date is before start date!"))
            return

        if self._always_save or self.changed is True:
            self.update_vevent()
            self.event.allday = self.startendeditor.allday
            self.event.increment_sequence()
            if self.event.etag is None:  # has not been saved before
                self.event.calendar = self.calendar_chooser.original_widget.active["name"]
                self.collection.insert(self.event)
            elif self.calendar_chooser.changed:
                self.collection.change_collection(self.event, self.calendar_chooser.active["name"])
            else:
                self.collection.update(self.event)

            self._save_callback(
                self.event.start_local,
                self.event.end_local,
                self.event.recurring or self.recurrenceeditor.changed,
            )
        self._abort_confirmed = False
        self.pane.window.backtrack()

    def keypress(self, size: tuple[int], key: str) -> str | None:
        if key in ["esc"] and self.changed and not self._abort_confirmed:
            self.pane.window.alert(("light red", "Unsaved changes! Hit ESC again to discard."))
            self._abort_confirmed = True
            return None
        else:
            self._abort_confirmed = False
        return_value = super().keypress(size, key)
        if key in self.pane._conf["keybindings"]["save"]:
            self.save(None)
            return None
        return return_value


WEEKDAYS = ["MO", "TU", "WE", "TH", "FR", "SA", "SU"]  # TODO use locale and respect weekdaystart


class WeekDaySelector(urwid.WidgetWrap):
    def __init__(self, startdt, selected_days) -> None:
        self._weekday_boxes = {day: urwid.CheckBox(day, state=False) for day in WEEKDAYS}
        weekday = startdt.weekday()
        self._weekday_boxes[WEEKDAYS[weekday]].state = True
        self.weekday_checks = NColumns([(7, self._weekday_boxes[wd]) for wd in WEEKDAYS])
        for day in selected_days:
            self._weekday_boxes[day].state = True
        urwid.WidgetWrap.__init__(self, self.weekday_checks)

    @property
    def days(self):
        days = [day.label for (day, _) in self.weekday_checks.contents if day.state]
        return days


class RecurrenceEditor(urwid.WidgetWrap):
    def __init__(self, rrule, conf, startdt) -> None:
        self._conf = conf
        self._startdt = startdt
        self._rrule = rrule
        self.repeat = bool(rrule)
        self._allow_edit = not self.repeat or self.check_understood_rrule(rrule)
        self.repeat_box = urwid.CheckBox(
            "Repeat: ",
            state=self.repeat,
            on_state_change=self.check_repeat,
        )

        if "UNTIL" in self._rrule:
            self._until = "Until"
        elif "COUNT" in self._rrule:
            self._until = "Repetitions"
        else:
            self._until = "Forever"

        recurrence = self._rrule["freq"][0].lower() if self._rrule else "weekly"
        self.recurrence_choice = CPadding(
            CAttrMap(
                Choice(
                    ["daily", "weekly", "monthly", "yearly"],
                    recurrence,
                    callback=self.rebuild,
                ),
                "popupper",
            ),
            align="center",
            left=2,
            right=2,
        )
        self.interval_edit = PositiveIntEdit(
            caption=("caption", "every:"),
            edit_text=str(self._rrule.get("INTERVAL", [1])[0]),
        )
        self.until_choice = CPadding(
            CAttrMap(
                Choice(
                    ["Forever", "Until", "Repetitions"],
                    self._until,
                    callback=self.rebuild,
                ),
                "popupper",
            ),
            align="center",
            left=2,
            right=2,
        )

        count = str(self._rrule.get("COUNT", [1])[0])
        self.repetitions_edit = PositiveIntEdit(edit_text=count)

        until = self._rrule.get("UNTIL", [None])[0]
        if until is None and isinstance(self._startdt, dt.datetime):
            until = self._startdt.date()
        elif until is None:
            until = self._startdt

        if isinstance(until, dt.datetime):
            until = until.date()
        self.until_edit = DateEdit(
            until,
            self._conf["locale"]["longdateformat"],
            lambda _: None,
            self._conf["locale"]["weeknumbers"],
            self._conf["locale"]["firstweekday"],
            self._conf["view"]["monthdisplay"],
        )

        self._rebuild_weekday_checks()
        self._rebuild_monthly_choice()
        self._pile = pile = NPile([urwid.Text("")])
        urwid.WidgetWrap.__init__(self, pile)
        self.rebuild()

    def _rebuild_monthly_choice(self):
        weekday, xth = get_weekday_occurrence(self._startdt)
        ords = {1: "st", 2: "nd", 3: "rd", 21: "st", 22: "nd", 23: "rd", 31: "st"}
        self._xth_weekday = f"on every {xth}{ords.get(xth, 'th')} {WEEKDAYS[weekday]}"
        self._xth_monthday = (
            f"on every {self._startdt.day}{ords.get(self._startdt.day, 'th')} of the month"
        )
        self.monthly_choice = Choice(
            [self._xth_monthday, self._xth_weekday],
            self._xth_monthday,
            callback=self.rebuild,
        )

    def _rebuild_weekday_checks(self):
        if self.recurrence_choice.active == "weekly":
            initial_days = self._rrule.get("BYDAY", [])
        else:
            initial_days = []
        self.weekday_checks = WeekDaySelector(self._startdt, initial_days)

    def update_startdt(self, startdt):
        self._startdt = startdt
        self._rebuild_weekday_checks()
        self._rebuild_monthly_choice()
        self.rebuild()

    @staticmethod
    def check_understood_rrule(rrule):
        """test if we can reproduce `rrule`."""
        keys = set(rrule.keys())
        freq = rrule.get("FREQ", [None])[0]
        unsupported_rrule_parts = {
            "BYSECOND",
            "BYMINUTE",
            "BYHOUR",
            "BYYEARDAY",
            "BYWEEKNO",
            "BYMONTH",
        }
        if keys.intersection(unsupported_rrule_parts):
            return False
        if len(rrule.get("BYMONTHDAY", [1])) > 1:
            return False
        # we don't support negative BYMONTHDAY numbers
        # don't need to check whole list, we only support one monthday anyway
        if rrule.get("BYMONTHDAY", [1])[0] < 1:
            return False
        if rrule.get("BYDAY", ["1"])[0][0] == "-":
            return False
        if rrule.get("BYSETPOS", [1])[0] != 1:
            return False
        if freq not in ["DAILY", "WEEKLY", "MONTHLY", "YEARLY"]:
            return False
        if "BYDAY" in keys and freq == "YEARLY":
            return False
        return True

    def check_repeat(self, checkbox, state):
        self.repeat = state
        self.rebuild()

    def _refill_contents(self, lines):
        while True:
            try:
                self._pile.contents.pop()
            except IndexError:
                break
        [self._pile.contents.append((line, ("pack", None))) for line in lines]

    def rebuild(self):
        old_focus_y = self._pile.focus_position
        if not self._allow_edit:
            self._rebuild_no_edit()
        elif self.repeat:
            self._rebuild_edit()
            self._pile.set_focus(old_focus_y)
        else:
            self._rebuild_edit_no_repeat()

    def _rebuild_no_edit(self):
        def _allow_edit(_):
            self._allow_edit = True
            self.rebuild()

        lines = [
            urwid.Text("We cannot reproduce this event's repetition rules."),
            urwid.Text("Editing the repetition rules will destroy the current rules."),
            urwid.Button("Edit anyway", on_press=_allow_edit),
        ]
        self._refill_contents(lines)
        self._pile.set_focus(2)

    def _rebuild_edit_no_repeat(self):
        lines = [NColumns([(13, self.repeat_box)])]
        self._refill_contents(lines)

    def _rebuild_edit(self):
        firstline = NColumns(
            [
                (13, self.repeat_box),
                (18, self.recurrence_choice),
                (13, self.interval_edit),
            ]
        )
        lines = [firstline]

        if self.recurrence_choice.active == "weekly":
            lines.append(self.weekday_checks)
        if self.recurrence_choice.active == "monthly":
            lines.append(self.monthly_choice)

        nextline = [(20, self.until_choice.original_widget)]
        if self.until_choice.active == "Until":
            nextline.append((24, self.until_edit))
        elif self.until_choice.active == "Repetitions":
            nextline.append((4, self.repetitions_edit))
        lines.append(NColumns(nextline))

        self._refill_contents(lines)

    @property
    def changed(self):
        # TODO this often gives false positives which leads to redraws of all
        # events shown
        return self._rrule != self.rrule()  # TODO do this properly

    def rrule(self):
        rrule = {}
        rrule["freq"] = [self.recurrence_choice.active]
        interval = int(self.interval_edit.get_edit_text())
        if interval != 1:
            rrule["interval"] = [interval]
        if rrule["freq"] == ["weekly"] and len(self.weekday_checks.days) > 1:
            rrule["byday"] = self.weekday_checks.days
        if rrule["freq"] == ["monthly"] and self.monthly_choice.active == self._xth_weekday:
            weekday, occurrence = get_weekday_occurrence(self._startdt)
            rrule["byday"] = [f"{occurrence}{WEEKDAYS[weekday]}"]
        if self.until_choice.active == "Until":
            if isinstance(self._startdt, dt.datetime):
                rrule["until"] = dt.datetime.combine(
                    self.until_edit.date,
                    self._startdt.time(),
                )
            else:
                rrule["until"] = self.until_edit.date
        elif self.until_choice.active == "Repetitions":
            rrule["count"] = int(self.repetitions_edit.get_edit_text())
        return rrule

    @property
    def active(self):
        if not self.repeat:
            return None
        else:
            return self.rrule()

    @active.setter
    def active(self, val):
        raise ValueError
        self.recurrence_choice.active = val


class ExportDialog(urwid.WidgetWrap):
    def __init__(self, this_func, abort_func, event) -> None:
        lines = []
        lines.append(urwid.Text("Export event as ICS file"))
        lines.append(urwid.Text(""))
        export_location = ExtendedEdit(
            caption="Location: ", edit_text=f"~/{event.summary.strip()}.ics"
        )
        lines.append(export_location)
        lines.append(urwid.Divider(" "))
        lines.append(
            CAttrMap(
                urwid.Button("Save", on_press=this_func, user_data=export_location),
                "button",
                "button focus",
            )
        )
        content = NPile(lines)
        urwid.WidgetWrap.__init__(self, urwid.LineBox(content))



# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


dark = [
    ("header", "white", "black"),
    ("footer", "white", "black"),
    ("line header", "black", "white", "bold"),
    ("alt header", "white", "", "bold"),
    ("bright", "dark blue", "white", "bold,standout"),
    ("list", "black", "white"),
    ("list focused", "white", "light blue", "bold"),
    ("edit", "black", "white"),
    ("edit focus", "white", "light blue", "bold"),
    ("button", "black", "dark cyan"),
    ("button focused", "white", "light blue", "bold"),
    ("reveal focus", "black", "light gray"),
    ("today focus", "white", "dark magenta"),
    (
        "today",
        "dark gray",
        "dark green",
    ),
    ("date header", "light gray", "black"),
    ("date header focused", "black", "white"),
    ("date header selected", "dark gray", "light gray"),
    ("dayname", "light gray", ""),
    ("monthname", "light gray", ""),
    ("weeknumber_right", "light gray", ""),
    ("alert", "white", "dark red"),
    ("mark", "white", "dark green"),
    ("frame", "white", "black"),
    ("frame focus", "light red", "black"),
    ("frame focus color", "dark blue", "black"),
    ("frame focus top", "dark magenta", "black"),
    ("eventcolumn", "", "", ""),
    ("eventcolumn focus", "", "", ""),
    ("calendar", "", "", ""),
    ("calendar focus", "", "", ""),
    ("editbx", "light gray", "dark blue"),
    ("editcp", "black", "light gray", "standout"),
    ("popupbg", "white", "black", "bold"),
    ("popupper", "white", "dark cyan"),
    ("caption", "white", "", "bold"),
]
light = [
    ("header", "black", "white"),
    ("footer", "black", "white"),
    ("line header", "black", "white", "bold"),
    ("alt header", "black", "", "bold"),
    ("bright", "dark blue", "white", "bold,standout"),
    ("list", "black", "white"),
    ("list focused", "white", "light blue", "bold"),
    ("edit", "black", "white"),
    ("edit focus", "white", "light blue", "bold"),
    ("button", "black", "dark cyan"),
    ("button focused", "white", "light blue", "bold"),
    ("reveal focus", "black", "dark cyan", "standout"),
    ("today focus", "white", "dark cyan", "standout"),
    ("today", "black", "light gray"),
    ("date header", "", "white"),
    ("date header focused", "white", "dark gray", "bold,standout"),
    ("date header selected", "dark gray", "light cyan"),
    ("dayname", "dark gray", "white"),
    ("monthname", "dark gray", "white"),
    ("weeknumber_right", "dark gray", "white"),
    ("alert", "white", "dark red"),
    ("mark", "white", "dark green"),
    ("frame", "dark gray", "white"),
    ("frame focus", "light red", "white"),
    ("frame focus color", "dark blue", "white"),
    ("frame focus top", "dark magenta", "white"),
    ("eventcolumn", "", "", ""),
    ("eventcolumn focus", "", "", ""),
    ("calendar", "", "", ""),
    ("calendar focus", "", "", ""),
    ("editbx", "light gray", "dark blue"),
    ("editcp", "black", "light gray", "standout"),
    ("popupbg", "white", "black", "bold"),
    ("popupper", "black", "light gray"),
    ("caption", "black", "", ""),
]

themes: dict[str, list[tuple[str, ...]]] = {"light": light, "dark": dark}



from .exceptions import InvalidSettingsError, NoConfigFile
from .settings import find_configuration_file, get_config

__all__ = [
    "InvalidSettingsError",
    "NoConfigFile",
    "find_configuration_file",
    "get_config",
]



# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
import logging
import sys

import click
import click_log

from . import __version__, khalendar
from .exceptions import FatalError
from .settings import InvalidSettingsError, NoConfigFile, get_config

logger = logging.getLogger("khal")
click_log.basic_config("khal")

days_option = click.option("--days", default=None, type=int, help="How many days to include.")
week_option = click.option("--week", "-w", help="Include all events in one week.", is_flag=True)
events_option = click.option("--events", default=None, type=int, help="How many events to include.")
dates_arg = click.argument("dates", nargs=-1)


def time_args(f):
    return dates_arg(events_option(week_option(days_option(f))))


def multi_calendar_select(ctx, include_calendars, exclude_calendars):
    if include_calendars and exclude_calendars:
        raise click.UsageError("Can't use both -a and -d.")

    selection = set()

    if include_calendars:
        for cal_name in include_calendars:
            if cal_name not in ctx.obj["conf"]["calendars"]:
                raise click.BadParameter(
                    f"Unknown calendar {cal_name}, run `khal printcalendars` "
                    "to get a list of all configured calendars."
                )

        selection.update(include_calendars)
    elif exclude_calendars:
        selection.update(ctx.obj["conf"]["calendars"].keys())
        for cal_name in exclude_calendars:
            if cal_name not in selection:
                raise click.BadParameter(
                    f"Unknown calendar {cal_name}, run `khal printcalendars` "
                    "to get a list of all configured calendars."
                )
            selection.discard(cal_name)

    return selection or None


def multi_calendar_option(f):
    a = click.option(
        "--include-calendar",
        "-a",
        multiple=True,
        metavar="CAL",
        help=("Include the given calendar. Can be specified multiple times."),
    )
    d = click.option(
        "--exclude-calendar",
        "-d",
        multiple=True,
        metavar="CAL",
        help=("Exclude the given calendar. Can be specified multiple times."),
    )

    return d(a(f))


def mouse_option(f):
    o = click.option(
        "--mouse/--no-mouse", is_flag=True, default=None, help="Disable mouse in interactive UI"
    )
    return o(f)


def _select_one_calendar_callback(ctx, option, calendar):
    if isinstance(calendar, tuple):
        if len(calendar) > 1:
            raise click.UsageError(
                'Can\'t use "--include-calendar" / "-a" more than once for this command.'
            )
        elif len(calendar) == 1:
            calendar = calendar[0]
    return _calendar_select_callback(ctx, option, calendar)


def _calendar_select_callback(ctx, option, calendar):
    if calendar and calendar not in ctx.obj["conf"]["calendars"]:
        raise click.BadParameter(
            f"Unknown calendar {calendar}, run `khal printcalendars` to get a "
            "list of all configured calendars."
        )
    return calendar


def calendar_option(f):
    return click.option("--calendar", "-a", metavar="CAL", callback=_calendar_select_callback)(f)


def global_options(f):
    def color_callback(ctx, option, value):
        ctx.color = value

    def logfile_callback(ctx, option, path):
        ctx.logfilepath = path

    config = click.option(
        "--config", "-c", help="The config file to use.", default=None, metavar="PATH"
    )
    color = click.option(
        "--color/--no-color",
        help=(
            "Use colored/uncolored output. Default is to only enable colors "
            "when not part of a pipe."
        ),
        expose_value=False,
        default=None,
        callback=color_callback,
    )

    logfile = click.option(
        "--logfile",
        "-l",
        help="The logfile to use [defaults to stdout]",
        type=click.Path(),
        callback=logfile_callback,
        default=None,
        expose_value=False,
        metavar="LOGFILE",
    )

    version = click.version_option(version=__version__)

    return logfile(config(color(version(f))))


def build_collection(conf, selection):
    """build and return a khalendar.CalendarCollection from the configuration"""
    try:
        props = {}
        for name, cal in conf["calendars"].items():
            if selection is None or name in selection:
                props[name] = {
                    "name": name,
                    "path": cal["path"],
                    "readonly": cal["readonly"],
                    "color": cal["color"],
                    "priority": cal["priority"],
                    "ctype": cal["type"],
                    "addresses": cal["addresses"] if "addresses" in cal else "",
                }
        collection = khalendar.CalendarCollection(
            calendars=props,
            color=conf["highlight_days"]["color"],
            locale=conf["locale"],
            dbpath=conf["sqlite"]["path"],
            hmethod=conf["highlight_days"]["method"],
            default_color=conf["highlight_days"]["default_color"],
            multiple=conf["highlight_days"]["multiple"],
            multiple_on_overflow=conf["highlight_days"]["multiple_on_overflow"],
            highlight_event_days=conf["default"]["highlight_event_days"],
        )
    except FatalError as error:
        logger.debug(error, exc_info=True)
        logger.fatal(error)
        sys.exit(1)

    collection._default_calendar_name = conf["default"]["default_calendar"]
    return collection


class _NoConfig:
    def __getitem__(self, key):
        logger.fatal(
            "Cannot find a config file. If you have no configuration file "
            "yet, you might want to run `khal configure`."
        )
        sys.exit(1)


def prepare_context(ctx, config):
    assert ctx.obj is None

    logger.debug("khal %s", __version__)
    try:
        conf = get_config(config)
    except NoConfigFile:
        conf = _NoConfig()
    except InvalidSettingsError:
        logger.info(
            "If your configuration file used to work, please have a "
            "look at the Changelog to see what changed."
        )
        sys.exit(1)
    else:
        logger.debug("Using config:")
        logger.debug(stringify_conf(conf))

    ctx.obj = {"conf_path": config, "conf": conf}


def stringify_conf(conf):
    # since we have only two levels of recursion, a recursive function isn't
    # really worth it
    out = []
    for key, value in conf.items():
        out.append(f"[{key}]")
        for subkey, subvalue in value.items():
            if isinstance(subvalue, dict):
                out.append(f"  [[{subkey}]]")
                for subsubkey, subsubvalue in subvalue.items():
                    out.append(f"    {subsubkey}: {subsubvalue}")
            else:
                out.append(f"  {subkey}: {subvalue}")
    return "\n".join(out)



# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""Contains a re-usable CalendarWidget for urwid.

if anything doesn't work as expected, please open an issue for khal
"""

import calendar
import datetime as dt
from collections.abc import Callable
from locale import LC_ALL, LC_TIME, getlocale, setlocale
from typing import Any, Literal, TypedDict

import urwid

from khal.utils import get_month_abbr_len


# Some custom types
class MarkType(TypedDict):
    date: dt.date
    pos: tuple[int, int]


OnPressType = dict[str, Callable[[dt.date, dt.date | None], str | None]]
GetStylesSignature = Callable[[dt.date, bool], str | tuple[str, str] | None]


setlocale(LC_ALL, "")


def getweeknumber(day: dt.date) -> int:
    """return iso week number for datetime.date object
    :param day: date
    :return: weeknumber
    """
    return dt.date.isocalendar(day)[1]


class DatePart(urwid.Text):
    """used in the Date widget (single digit)"""

    def __init__(self, digit: str) -> None:
        super().__init__(digit)

    @classmethod
    def selectable(cls: type) -> bool:
        return True

    def keypress(self, size: tuple[int], key: str) -> str:
        return key

    def get_cursor_coords(self, size: tuple[int]) -> tuple[int, int]:
        return 1, 0

    def render(self, size: tuple[int], focus: bool = False) -> urwid.Canvas:
        canv = super().render(size, focus)
        if focus:
            canv = urwid.CompositeCanvas(canv)
            canv.cursor = 1, 0
        return canv


class Date(urwid.WidgetWrap):
    """used in the main calendar for dates (a number)"""

    def __init__(self, date: dt.date, get_styles: GetStylesSignature) -> None:
        dstr = str(date.day).rjust(2)
        self.halves = [
            urwid.AttrMap(DatePart(dstr[:1]), None, None),
            urwid.AttrMap(DatePart(dstr[1:]), None, None),
        ]
        self.date = date
        self._get_styles = get_styles
        super().__init__(urwid.Columns(self.halves))

    def set_styles(self, styles: None | str | tuple[str, str]) -> None:
        """If single string, sets the same style for both halves, if two
        strings, sets different style for each half.
        """
        if type(styles) is tuple:
            self.halves[0].set_attr_map({None: styles[0]})
            self.halves[1].set_attr_map({None: styles[1]})
            self.halves[0].set_focus_map({None: styles[0]})
            self.halves[1].set_focus_map({None: styles[1]})
        else:
            self.halves[0].set_attr_map({None: styles})
            self.halves[1].set_attr_map({None: styles})
            self.halves[0].set_focus_map({None: styles})
            self.halves[1].set_focus_map({None: styles})

    def reset_styles(self, focus: bool = False) -> None:
        self.set_styles(self._get_styles(self.date, focus))

    @property
    def marked(self) -> bool:
        return "mark" in [self.halves[0].attr_map[None], self.halves[1].attr_map[None]]

    @classmethod
    def selectable(cls) -> bool:
        return True

    def keypress(self, _: Any, key: str) -> str:
        return key


class DateCColumns(urwid.Columns):
    """container for one week worth of dates
    which are horizontally aligned

    TODO: rename, awful name

    focus can only move away by pressing 'TAB',
    calls 'on_date_change' on every focus change (see below for details)
    """

    # TODO only call on_date_change when we change our date ourselves,
    # not if it gets changed by an (external) call to set_focus_date()

    def __init__(
        self,
        widget_list,
        on_date_change: Callable[[dt.date], None],
        on_press: OnPressType,
        keybindings: dict[str, list[str]],
        get_styles: GetStylesSignature,
        **kwargs,
    ) -> None:
        self.on_date_change = on_date_change
        self.on_press = on_press
        self.keybindings = keybindings
        self.get_styles = get_styles
        self._init: bool = True
        super().__init__(widget_list, **kwargs)

    def __repr__(self) -> str:
        return f"<DateCColumns from {self[1].date} to {self[7].date}>"

    def _clear_cursor(self) -> None:
        old_pos: int = self.focus_position
        self.contents[old_pos][0].set_styles(self.get_styles(self.contents[old_pos][0].date, False))

    @property
    def focus_position(self) -> int:
        """returns the current focus position"""
        return urwid.Columns.focus_position.fget(self)

    @focus_position.setter
    def focus_position(self, position: int) -> None:
        """calls on_date_change before setting super().focus_position"""
        # do not call when building up the interface, lots of potentially
        # expensive calls made here
        if self._init:
            self._init = False
        else:
            self._clear_cursor()
            self.contents[position][0].set_styles(
                self.get_styles(self.contents[position][0].date, True)
            )
            self.on_date_change(self.contents[position][0].date)
        urwid.Columns.focus_position.fset(self, position)

    def set_focus_date(self, a_date: dt.date) -> None:
        for num, day in enumerate(self.contents[1:8], 1):
            if day[0].date == a_date:
                self.focus_position = num
                return None
        raise ValueError(f"{a_date} not found in this week")

    def get_date_column(self, a_date: dt.date) -> int:
        """return the column `a_date` is in, raises ValueError if `a_date`
        cannot be found
        """
        for num, day in enumerate(self.contents[1:8], 1):
            if day[0].date == a_date:
                return num
        raise ValueError(f"{a_date} not found in this week")

    def keypress(self, size: tuple[int], key: str) -> str:
        """only leave calendar area on pressing 'tab' or 'enter'"""

        if key in self.keybindings["left"]:
            key = "left"
        elif key in self.keybindings["up"]:
            key = "up"
        elif key in self.keybindings["right"]:
            key = "right"
        elif key in self.keybindings["down"]:
            key = "down"

        exit_row = False  # set this, if we are leaving the current row
        old_pos = self.focus_position

        key = super().keypress(size, key)

        # make sure we don't leave the calendar
        if old_pos == 7 and key == "right":
            self.focus_position = 1
            exit_row = True
            key = "down"
        elif old_pos == 1 and key == "left":
            self.focus_position = 7
            exit_row = True
            key = "up"
        elif key in self.keybindings["view"]:  # TODO make this more generic
            self.focus_position = old_pos
            key = "right"
        elif key in ["up", "down"]:
            exit_row = True

        if exit_row:
            self._clear_cursor()

        return key


class CListBox(urwid.ListBox):
    """our custom version of ListBox containing a CalendarWalker instance

    it should contain a `CalendarWalker` instance which it autoextends on
    rendering, if needed"""

    def __init__(self, walker: "CalendarWalker") -> None:
        self._init: bool = True
        self.keybindings = walker.keybindings
        self.on_press = walker.on_press
        self._marked: MarkType | None = None
        self._pos_old: tuple[int, int] | None = None
        self.body: CalendarWalker
        super().__init__(walker)

    @property
    def focus_position(self) -> int:
        return super().focus_position

    @focus_position.setter
    def focus_position(self, position: int) -> None:
        super().set_focus(position)

    def render(self, size: tuple[int], focus: bool = False) -> urwid.Canvas:
        while "bottom" in self.ends_visible(size):
            self.body._autoextend()
        if self._init:
            self.set_focus_valign("middle")
            self._init = False

        return super().render(size, focus)

    def mouse_event(self, *args):
        _size, event, button, _col, _row, _focus = args

        if event == "mouse press" and button == 1:
            self.focus.focus.set_styles(self.focus.get_styles(self.body.focus_date, False))
        return super().mouse_event(*args)

    def _date(self, row: int, column: int) -> dt.date:
        """return the date at row `row` and  column `column`"""
        return self.body[row].contents[column][0].date

    def _unmark_one(self, row: int, column: int) -> None:
        """remove attribute *mark* from the date at row `row` and column `column`
        returning it to the attributes defined by self._get_color()
        """
        self.body[row].contents[column][0].reset_styles()

    def _mark_one(self, row: int, column: int) -> None:
        """set attribute *mark* on the date at row `row` and column `column`"""
        self.body[row].contents[column][0].set_styles("mark")

    def _mark(self, a_date: dt.date | None = None) -> None:
        """make sure everything between the marked entry and `a_date`
        is visually marked, and nothing else"""

        assert self._marked is not None

        if a_date is None:
            a_date = self.body.focus_date

        def toggle(row: int, column: int) -> None:
            """toggle the mark attribute on the date at row `row` and column
            `column`"""
            if self.body[row].contents[column][0].marked:
                self._mark_one(row, column)
            else:
                self._unmark_one(row, column)

        start = min(self._marked["pos"][0], self.focus_position) - 2
        stop = max(self._marked["pos"][0], self.focus_position) + 2
        for row in range(start, stop):
            for col in range(1, 8):
                if a_date > self._marked["date"]:
                    if self._marked["date"] <= self._date(row, col) <= a_date:
                        self._mark_one(row, col)
                    else:
                        self._unmark_one(row, col)
                else:
                    if self._marked["date"] >= self._date(row, col) >= a_date:
                        self._mark_one(row, col)
                    else:
                        self._unmark_one(row, col)

            toggle(self.focus_position, self.focus.focus_col)
        self._pos_old = self.focus_position, self.focus.focus_col

    def _unmark_all(self) -> None:
        """remove attribute *mark* from all dates"""
        if self._marked and self._pos_old:
            start = min(self._marked["pos"][0], self.focus_position, self._pos_old[0])
            end = max(self._marked["pos"][0], self.focus_position, self._pos_old[0]) + 1
            for row in range(start, end):
                for col in range(1, 8):
                    self._unmark_one(row, col)

    def set_focus_date(self, a_day: dt.date) -> None:
        """set focus to the date `a_day`"""
        self.focus.focus.set_styles(self.focus.get_styles(self.body.focus_date, False))
        if self._marked:
            self._unmark_all()
            self._mark(a_day)
        self.body.set_focus_date(a_day)

    def keypress(self, size: bool, key: str) -> str | None:
        if key in self.keybindings["mark"] + ["esc"] and self._marked:
            self._unmark_all()
            self._marked = None
            return None
        if key in self.keybindings["mark"]:
            self._marked = {
                "date": self.body.focus_date,
                "pos": (self.focus_position, self.focus.focus_col),
            }
        if self._marked and key in self.keybindings["other"]:
            row, col = self._marked["pos"]
            self._marked = {
                "date": self.body.focus_date,
                "pos": (self.focus_position, self.focus.focus_col),
            }
            self.focus.focus_col = col
            self.focus_position = row
        if key in self.on_press:
            if self._marked:
                start = min(self.body.focus_date, self._marked["date"])
                end = max(self.body.focus_date, self._marked["date"])
            else:
                start = self.body.focus_date
                end = None
            return self.on_press[key](start, end)
        if key in self.keybindings["today"] + ["page down", "page up"]:
            # reset colors of currently focused Date widget
            self.focus.focus.set_styles(self.focus.get_styles(self.body.focus_date, False))
        if key in self.keybindings["today"]:
            self.set_focus_date(dt.date.today())
            self.set_focus_valign(("relative", 10))

        key = super().keypress(size, key)
        if self._marked:
            self._mark()
        return key


class CalendarWalker(urwid.SimpleFocusListWalker):
    def __init__(
        self,
        on_date_change: Callable[[dt.date], None],
        on_press: dict[str, Callable[[dt.date, dt.date | None], str | None]],
        keybindings: dict[str, list[str]],
        get_styles: GetStylesSignature,
        firstweekday: int = 0,
        weeknumbers: Literal["left", "right", False] = False,
        monthdisplay: Literal["firstday", "firstfullweek"] = "firstday",
        initial: dt.date | None = None,
    ) -> None:
        self.firstweekday = firstweekday
        self.weeknumbers = weeknumbers
        self.monthdisplay = monthdisplay
        self.on_date_change = on_date_change
        self.on_press = on_press
        self.keybindings = keybindings
        self.get_styles = get_styles
        self.reset(initial)

    def reset(self, initial: dt.date | None = None) -> None:
        if initial is None:
            initial = dt.date.today()
        weeks = self._construct_month(initial.year, initial.month)
        urwid.SimpleFocusListWalker.__init__(self, weeks)

    def set_focus(self, position: int) -> None:
        """set focus by item number"""
        while position >= len(self) - 1:
            self._autoextend()
        while position <= 0:
            no_additional_weeks = self._autoprepend()
            position += no_additional_weeks
        urwid.SimpleFocusListWalker.set_focus(self, position)

    def days_to_next_already_loaded(self, day: dt.date) -> int:
        """return the number of weeks from the focus to the next week that is already loaded"""
        if len(self) == 0:
            return 0
        elif self.earliest_date <= day <= self.latest_date:
            return 0
        elif day <= self.earliest_date:
            return (self.earliest_date - day).days
        elif self.latest_date <= day:
            return (day - self.latest_date).days
        else:
            raise ValueError("This should not happen")

    @property
    def focus_date(self) -> dt.date:
        """return the date the focus is currently set to"""
        return self[self.focus].focus.date

    def set_focus_date(self, a_day: dt.date) -> None:
        """set the focus to `a_day`"""
        if self.days_to_next_already_loaded(a_day) > 200:  # arbitrary number
            self.reset(a_day)
        row, column = self.get_date_pos(a_day)
        self.set_focus(row)
        self[self.focus].focus_position = column

    @property
    def earliest_date(self) -> dt.date:
        """return earliest day that is already loaded into the CalendarWidget"""
        return self[0][1].date

    @property
    def latest_date(self) -> dt.date:
        """return latest day that is already loaded into the CalendarWidget"""
        return self[-1][7].date

    def reset_styles_range(self, min_date: dt.date, max_date: dt.date) -> None:
        """reset styles for all (displayed) dates between min_date and max_date"""
        minr, _minc = self.get_date_pos(max(min_date, self.earliest_date))
        maxr, _maxc = self.get_date_pos(min(max_date, self.latest_date))
        focus_pos = self.focus, self[self.focus].focus_col

        for row in range(minr, maxr + 1):
            for column in range(1, 8):
                focus = (row, column) == focus_pos
                self[row][column].reset_styles(focus)

    def get_date_pos(self, a_day: dt.date) -> tuple[int, int]:
        """get row and column where `a_day` is located"""
        # rough estimate of difference in lines, i.e. full weeks, we might be
        # off by as much as one week though
        week_diff = int((self.focus_date - a_day).days / 7)
        new_focus = self.focus - week_diff
        # in case new_focus is 1 we will later try set the focus to 0 which
        # will lead to an autoprepend which will f*ck up our estimation,
        # therefore better autoprepending anyway, even if it might not be
        # necessary
        while new_focus <= 1:
            self._autoprepend()
            week_diff = int((self.focus_date - a_day).days / 7)
            new_focus = self.focus - week_diff
        for offset in [0, -1, 1]:  # we might be off by a week
            row = new_focus + offset
            try:
                while row >= len(self):
                    self._autoextend()
                column = self[row].get_date_column(a_day)
                return row, column
            except (ValueError, IndexError):
                pass
        # we didn't find the date we were looking for...
        raise ValueError("something is wrong")

    def _autoextend(self) -> None:
        """appends the next month"""
        date_last_month = self[-1][1].date  # a date from the last month
        last_month = date_last_month.month
        last_year = date_last_month.year
        month = last_month % 12 + 1
        year = last_year if not last_month == 12 else last_year + 1
        weeks = self._construct_month(year, month, clean_first_row=True)
        self.extend(weeks)

    def _autoprepend(self) -> int:
        """prepends the previous month

        :returns: number of weeks prepended
        """
        try:
            date_first_month = self[0][-1].date  # a date from the first month
        except AttributeError:
            # rightmost column is weeknumber
            date_first_month = self[0][-2].date
        first_month = date_first_month.month
        first_year = date_first_month.year
        if first_month == 1:
            month = 12
            year = first_year - 1
        else:
            month = first_month - 1
            year = first_year
        weeks = self._construct_month(year, month, clean_last_row=True)
        weeks.reverse()
        for one in weeks:
            self.insert(0, one)
        return len(weeks)

    def _construct_week(self, week: list[dt.date]) -> DateCColumns:
        """
        constructs a CColumns week from a week of datetime.date objects. Also
        prepends the month name if the first day of the month is included in
        that week.

        :param week: list of datetime.date objects
        :returns: the week as an CColumns object
        """
        if self.monthdisplay == "firstday" and 1 in (day.day for day in week):
            month_name = calendar.month_abbr[week[-1].month].ljust(4)
            attr = "monthname"
        elif self.monthdisplay == "firstfullweek" and week[0].day <= 7:
            month_name = calendar.month_abbr[week[-1].month].ljust(4)
            attr = "monthname"
        elif self.weeknumbers == "left":
            month_name = f" {getweeknumber(week[0]):2} "
            attr = "weeknumber_left"
        else:
            month_name = "    "
            attr = None

        this_week: list[tuple[int, Date | urwid.AttrMap]]
        this_week = [(get_month_abbr_len(), urwid.AttrMap(urwid.Text(month_name), attr))]
        for _number, day in enumerate(week):
            new_date = Date(day, self.get_styles)
            this_week.append((2, new_date))
            new_date.set_styles(self.get_styles(new_date.date, False))
        if self.weeknumbers == "right":
            this_week.append(
                (2, urwid.AttrMap(urwid.Text(f"{getweeknumber(week[0]):2}"), "weeknumber_right"))
            )

        week = DateCColumns(
            this_week,
            on_date_change=self.on_date_change,
            on_press=self.on_press,
            keybindings=self.keybindings,
            dividechars=1,
            get_styles=self.get_styles,
        )
        return week

    def _construct_month(
        self,
        year: int = dt.date.today().year,
        month: int = dt.date.today().month,
        clean_first_row: bool = False,
        clean_last_row: bool = False,
    ) -> list[DateCColumns]:
        """construct one month of DateCColumns

        :param year: the year this month is set in
        :param month: the number of the month to be constructed
        :param clean_first_row: if set, makes sure that the first element
           returned is completely in `month` and not partly in the one before
           (which might lead to that line occurring twice
        :param clean_last_row: if set, makes sure that the last element returned
            is completely in `month` and not partly in the one after (which
            might lead to that line occurring twice)
        :returns: list of DateCColumns and the number of the list element which
                  contains today (or None if it isn't in there)
        """

        plain_weeks = calendar.Calendar(self.firstweekday).monthdatescalendar(year, month)
        weeks = []
        for _number, week in enumerate(plain_weeks):
            weeks.append(self._construct_week(week))
        if clean_first_row and weeks[0][1].date.month != weeks[0][7].date.month:
            return weeks[1:]
        elif clean_last_row and weeks[-1][1].date.month != weeks[-1][7].date.month:
            return weeks[:-1]
        else:
            return weeks


class CalendarWidget(urwid.WidgetWrap):
    def __init__(
        self,
        on_date_change: Callable[[dt.date], None],
        keybindings: dict[str, list[str]],
        on_press: OnPressType | None = None,
        firstweekday: int = 0,
        weeknumbers: Literal["left", "right", False] = False,
        monthdisplay: Literal["firstday", "firstfullweek"] = "firstday",
        get_styles: GetStylesSignature | None = None,
        initial: dt.date | None = None,
    ) -> None:
        """A calendar widget that can be used in urwid applications

        :param on_date_change: a function that is called every time the selected
            date is changed with the newly selected date as an argument
        :param keybindings: bind keys to specific functionionality, keys are
            the available commands, values are lists of keys that should be
            bound to those commands. See below for the defaults.
            Available commands:
                'left', 'right', 'up', 'down': move cursor in that direction
                'today': refocus on today
                'mark': toggles selection mode
                'other': toggles between selecting the earlier and the later end
                         of a selection
                'view': returns the key `right` to the widget containing the
                        CalendarWidget
        :param on_press: dictonary of functions that are called when the key is
            pressed and is not already bound to one of the internal functions
            via `keybindings`. These functions must accept two arguments, in
            normal mode the first argument is the currently selected date
            (datetime.date) and the second is `None`. When a date range is
            selected, the first argument is the earlier, the second argument
            is the later date. The function's return values are interpreted as
            pressed keys, which are handed to the widget containing the
            CalendarWidget.
        :param firstweekday: the first day of the week, 0 for Monday, 6 for
        :param weeknumbers: display weeknumbers on the left or right side of
           the calendar.
        :param monthdisplay: display the month name in the row of the 1st of the
           month or in the first row that only contains days of the current month.
        :param get_styles: a function that returns a list of styles for a given date
        :param initial: the date that is selected when the widget is first rendered
        """
        if initial is None:
            self._initial = dt.date.today()
        else:
            self._initial = initial

        if on_press is None:
            on_press = {}

        default_keybindings: dict[str, list[str]] = {
            "left": ["left"],
            "down": ["down"],
            "right": ["right"],
            "up": ["up"],
            "today": ["t"],
            "view": [],
            "mark": ["v"],
            "other": ["%"],
        }

        default_keybindings.update(keybindings)
        calendar.setfirstweekday(firstweekday)

        try:
            mylocale: str = ".".join(getlocale(LC_TIME))  # type: ignore
        except TypeError:  # language code and encoding may be None
            mylocale = "C"

        _calendar = calendar.LocaleTextCalendar(firstweekday, mylocale)  # type: ignore
        weekheader = _calendar.formatweekheader(2)
        dnames = weekheader.split(" ")

        def _get_styles(date: dt.date, focus: bool) -> str | None:
            if focus:
                if date == dt.date.today():
                    return "today focus"
                else:
                    return "reveal focus"
            else:
                if date == dt.date.today():
                    return "today"
                else:
                    return None

        if get_styles is None:
            get_styles = _get_styles

        if weeknumbers == "right":
            dnames.append("#w")
        month_names_length = get_month_abbr_len()
        cnames = urwid.Columns(
            [(month_names_length, urwid.Text(" " * month_names_length))]
            + [(2, urwid.AttrMap(urwid.Text(name), "dayname")) for name in dnames],
            dividechars=1,
        )
        self.walker = CalendarWalker(
            on_date_change=on_date_change,
            on_press=on_press,
            keybindings=default_keybindings,
            firstweekday=firstweekday,
            weeknumbers=weeknumbers,
            monthdisplay=monthdisplay,
            get_styles=get_styles,
            initial=self._initial,
        )
        self.box = CListBox(self.walker)
        frame = urwid.Frame(self.box, header=cnames)
        urwid.WidgetWrap.__init__(self, frame)
        self.set_focus_date(self._initial)

    def focus_today(self) -> None:
        self.set_focus_date(dt.date.today())

    def reset_styles_range(self, min_date: dt.date, max_date: dt.date) -> None:
        self.walker.reset_styles_range(min_date, max_date)

    @classmethod
    def selectable(cls) -> bool:
        return True

    @property
    def focus_date(self) -> dt.date:
        return self.walker.focus_date

    def set_focus_date(self, a_day: dt.date) -> None:
        """set the focus to `a_day`"""
        self.box.set_focus_date(a_day)



# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


"""
CalendarCollection should enable modifying and querying a collection of
calendars. Each calendar is defined by the contents of a vdir, but uses an
SQLite db for caching (see backend if you're interested).
"""

import datetime as dt
import itertools
import logging
import os
import os.path
from collections.abc import Iterable

from khal.custom_types import CalendarConfiguration, EventCreationTypes, LocaleConfiguration
from khal.icalendar import new_vevent

from . import backend
from .event import Event
from .exceptions import (
    DuplicateUid,
    EtagMissmatch,
    NonUniqueUID,
    ReadOnlyCalendarError,
    UnsupportedFeatureError,
    UpdateFailed,
)
from .vdir import (
    AlreadyExistingError,
    CollectionNotFoundError,
    Vdir,
    WrongEtagError,
    get_etag_from_file,
)

logger = logging.getLogger("khal")


class CalendarCollection:
    """CalendarCollection allows access to various calendars stored in vdirs

    all calendars are cached in an sqlitedb for performance reasons"""

    def __init__(
        self,
        calendars: dict[str, CalendarConfiguration],
        hmethod: str = "fg",
        default_color: str = "",
        multiple: str = "",
        multiple_on_overflow: bool = False,
        color: str = "",
        priority: int = 10,
        highlight_event_days: bool = False,
        locale: LocaleConfiguration | None = None,
        dbpath: str | None = None,
    ) -> None:
        assert locale
        assert dbpath is not None
        assert calendars is not None

        self._calendars: dict[str, CalendarConfiguration] = calendars
        self._default_calendar_name: str | None = None
        self._storages: dict[str, Vdir] = {}
        file_ext: str

        for name, calendar in self._calendars.items():
            ctype = calendar.get("ctype", "calendar")
            if ctype == "calendar":
                file_ext = ".ics"
            elif ctype == "birthdays":
                file_ext = ".vcf"
            else:
                raise ValueError("ctype must be either `calendar` or `birthdays`")
            try:
                self._storages[name] = Vdir(calendar["path"], file_ext)
            except CollectionNotFoundError:
                os.makedirs(calendar["path"])
                logger.info(f"created non-existing vdir {calendar['path']}")
                self._storages[name] = Vdir(calendar["path"], file_ext)

        self.hmethod = hmethod
        self.default_color = default_color
        self.multiple = multiple
        self.multiple_on_overflow = multiple_on_overflow
        self.color = color
        self.priority = priority
        self.highlight_event_days = highlight_event_days
        self._locale = locale
        self._backend = backend.SQLiteDb(self.names, dbpath, self._locale)
        self._last_ctags: dict[str, str] = {}
        self.update_db()

    @property
    def writable_names(self) -> list[str]:
        return [c for c in self._calendars if not self._calendars[c].get("readonly", False)]

    @property
    def calendars(self) -> Iterable[CalendarConfiguration]:
        return self._calendars.values()

    @property
    def names(self) -> Iterable[str]:
        return self._calendars.keys()

    @property
    def default_calendar_name(self) -> str | None:
        return self._default_calendar_name

    @default_calendar_name.setter
    def default_calendar_name(self, default: str) -> None:
        if default is None:
            self._default_calendar_name = default
        elif default not in self.names:
            raise ValueError(f"Unknown calendar: {default}")

        readonly = self._calendars[default].get("readonly", False)

        if not readonly:
            self._default_calendar_name = default
        else:
            raise ValueError(f'Calendar "{default}" is read-only and cannot be used as default')

    def _local_ctag(self, calendar: str) -> str:
        return get_etag_from_file(self._calendars[calendar]["path"])

    def get_floating(self, start: dt.datetime, end: dt.datetime) -> Iterable[Event]:
        for args in self._backend.get_floating(start, end):
            yield self._construct_event(*args)

    def get_localized(self, start: dt.datetime, end: dt.datetime) -> Iterable[Event]:
        for args in self._backend.get_localized(start, end):
            yield self._construct_event(*args)

    def get_events_on(self, day: dt.date) -> Iterable[Event]:
        """return all events on `day`"""
        start = dt.datetime.combine(day, dt.time.min)
        end = dt.datetime.combine(day, dt.time.max)
        floating_events = self.get_floating(start, end)
        localize = self._locale["local_timezone"].localize
        localized_events = self.get_localized(localize(start), localize(end))
        return itertools.chain(localized_events, floating_events)

    def get_calendars_on(self, day: dt.date) -> list[str]:
        start = dt.datetime.combine(day, dt.time.min)
        end = dt.datetime.combine(day, dt.time.max)
        localize = self._locale["local_timezone"].localize
        calendars = itertools.chain(
            self._backend.get_floating_calendars(start, end),
            self._backend.get_localized_calendars(localize(start), localize(end)),
        )
        return list(set(calendars))

    def update(self, event: Event) -> None:
        """update `event` in vdir and db"""
        assert event.etag is not None
        assert event.calendar is not None
        assert event.href is not None
        assert event.raw is not None
        if self._calendars[event.calendar]["readonly"]:
            raise ReadOnlyCalendarError()
        with self._backend.at_once():
            event.etag = self._storages[event.calendar].update(event.href, event, event.etag)
            self._backend.update(event.raw, event.href, event.etag, calendar=event.calendar)
            self._backend.set_ctag(self._local_ctag(event.calendar), calendar=event.calendar)

    def force_update(self, event: Event, collection: str | None = None) -> None:
        """update `event` even if an event with the same uid/href already exists"""
        href: str
        calendar = collection if collection is not None else event.calendar
        assert calendar is not None
        if self._calendars[calendar]["readonly"]:
            raise ReadOnlyCalendarError()

        with self._backend.at_once():
            try:
                href, etag = self._storages[calendar].upload(event)
            except AlreadyExistingError as error:
                href = error.existing_href
                _, etag = self._storages[calendar].get(href)
                etag = self._storages[calendar].update(href, event, etag)
            self._backend.update(event.raw, href, etag, calendar=calendar)
            self._backend.set_ctag(self._local_ctag(calendar), calendar=calendar)

    def insert(self, event: Event, collection: str | None = None) -> None:
        """Insert a new event to the vdir and the database

        The event will get a new href and etag properties. If ``collection`` is
        ``None``, then ``event.calendar`` must be defined.

        :param event: the event to be inserted.
        """
        # TODO FIXME not all `event`s are actually of type Event, we also uptade
        # with vdir.Items. Those don't have an .href or .etag property which we
        # than attach anyway. Works, but pretty ugly and any type checker will
        # complain.
        calendar = collection if collection is not None else event.calendar
        assert calendar is not None
        if hasattr(event, "etag"):
            assert not event.etag
        if self._calendars[calendar]["readonly"]:
            raise ReadOnlyCalendarError()

        with self._backend.at_once():
            try:
                event.href, event.etag = self._storages[calendar].upload(event)
            except AlreadyExistingError as Error:
                href = getattr(Error, "existing_href", None)
                raise DuplicateUid(href)
            self._backend.update(event.raw, event.href, event.etag, calendar=calendar)
            self._backend.set_ctag(self._local_ctag(calendar), calendar=calendar)

    def delete(self, href: str, etag: str | None, calendar: str) -> None:
        """Delete an event specified by `href` from `calendar`"""
        if self._calendars[calendar]["readonly"]:
            raise ReadOnlyCalendarError()
        try:
            self._storages[calendar].delete(href, etag)
        except WrongEtagError:
            raise EtagMissmatch()
        self._backend.delete(href, calendar=calendar)

    def delete_instance(
        self,
        href: str,
        etag: str | None,
        calendar: str,
        rec_id: dt.datetime,
    ) -> Event:
        """Delete a recurrence instance from an event specified by `href` from `calendar`

        returns the updated event
        """
        if self._calendars[calendar]["readonly"]:
            raise ReadOnlyCalendarError()
        event = self.get_event(href, calendar)
        if etag and etag != event.etag:
            raise EtagMissmatch()

        event.delete_instance(rec_id)
        self.update(event)
        return event

    def get_event(self, href: str, calendar: str) -> Event:
        """get an event by its href from the datatbase"""
        event_str, etag = self._backend.get_with_etag(href, calendar)
        return self._construct_event(event_str, etag=etag, href=href, calendar=calendar)

    def _construct_event(
        self,
        item: str,
        href: str,
        start: dt.datetime | dt.date | None = None,
        end: dt.datetime | dt.date | None = None,
        ref: str = "PROTO",
        etag: str | None = None,
        calendar: str | None = None,
    ) -> Event:
        assert calendar is not None
        event = Event.fromString(
            item,
            locale=self._locale,
            href=href,
            calendar=calendar,
            etag=etag,
            start=start,
            end=end,
            ref=ref,
            color=self._calendars[calendar]["color"],
            readonly=self._calendars[calendar]["readonly"],
            addresses=self._calendars[calendar]["addresses"],
        )
        return event

    def change_collection(self, event: Event, new_collection: str) -> None:
        """Moves `event` to a new collection (calendar)"""
        href, etag, calendar = event.href, event.etag, event.calendar
        event.etag = None
        self.insert(event, new_collection)
        assert href is not None
        assert calendar is not None
        self.delete(href, etag, calendar=calendar)

    def create_event_from_ics(
        self,
        ical: str,
        calendar_name: str,
        etag: str | None = None,
        href: str | None = None,
    ) -> Event:
        """creates and returns (but does not insert) a new event from ical
        string"""
        calendar = calendar_name or self.writable_names[0]
        return Event.fromString(ical, locale=self._locale, calendar=calendar, etag=etag, href=href)

    def create_event_from_dict(
        self,
        event_dict: EventCreationTypes,
        calendar_name: str | None = None,
    ) -> Event:
        """Creates an Event from the method's arguments"""
        vevent = new_vevent(locale=self._locale, **event_dict)
        calendar_name = calendar_name or self.default_calendar_name or self.writable_names[0]
        assert calendar_name is not None
        return self.create_event_from_ics(vevent.to_ical(), calendar_name)

    def update_db(self) -> None:
        """update the db from the vdir,

        should be called after every change to the vdir
        """
        for calendar in self._calendars:
            if self._needs_update(calendar, remember=True):
                self._db_update(calendar)

    def needs_update(self) -> bool:
        """Check if you need to call update_db.

        This could either be the case because the vdirs were changed externally,
        or another instance of khal updated the caching db already.
        """
        # TODO is it a good idea to munch both use cases together?
        # in case another instance of khal has updated the db, we only need
        # to get new events, but # update_db() takes potentially a long time to return
        # but then the code (in ikhal's refresh code) would need to look like
        # this:
        #
        # update_ui = False
        # if collection.needs_update():
        #   collection.update_db()
        #   update_ui = True
        # if collection.needs_refresh() or update_ui:
        #   do_the_update()
        #
        # and the API would be made even uglier than it already is...
        for calendar in self._calendars:
            if self._needs_update(calendar) or self._last_ctags[calendar] != self._local_ctag(
                calendar
            ):
                return True
        return False

    def _needs_update(self, calendar: str, remember: bool = False) -> bool:
        """checks if the db for the given calendar needs an update"""
        local_ctag = self._local_ctag(calendar)
        if remember:
            self._last_ctags[calendar] = local_ctag
        return local_ctag != self._backend.get_ctag(calendar)

    def _db_update(self, calendar: str) -> None:
        """implements the actual db update on a per calendar base"""
        local_ctag = self._local_ctag(calendar)
        db_hrefs = {href for href, etag in self._backend.list(calendar)}
        storage_hrefs: set[str] = set()
        bdays = self._calendars[calendar].get("ctype") == "birthdays"

        with self._backend.at_once():
            for href, etag in self._storages[calendar].list():
                storage_hrefs.add(href)
                db_etag = self._backend.get_etag(href, calendar=calendar)
                if etag != db_etag:
                    logger.debug(f"Updating {href} because {etag} != {db_etag}")
                    self._update_vevent(href, calendar=calendar)
            for href in db_hrefs - storage_hrefs:
                if bdays:
                    for sh in storage_hrefs:
                        if href.startswith(sh):
                            break
                    else:
                        self._backend.delete(href, calendar=calendar)
                else:
                    self._backend.delete(href, calendar=calendar)
            self._backend.set_ctag(local_ctag, calendar=calendar)
            self._last_ctags[calendar] = local_ctag

    def _update_vevent(self, href: str, calendar: str) -> bool:
        """should only be called during db_update, only updates the db,
        does not check for readonly"""
        event, etag = self._storages[calendar].get(href)
        try:
            if self._calendars[calendar].get("ctype") == "birthdays":
                update = self._backend.update_vcf_dates
            else:
                update = self._backend.update
            update(event.raw, href=href, etag=etag, calendar=calendar)
            return True
        except Exception as e:
            if not isinstance(e, UpdateFailed | UnsupportedFeatureError | NonUniqueUID):
                logger.exception("Unknown exception happened.")
            logger.warning(
                f"Skipping {calendar}/{href}: {e!s}\nThis event will not be available in khal."
            )
            return False

    def search(self, search_string: str) -> Iterable[Event]:
        """search for the db for events matching `search_string`"""
        return (self._construct_event(*args) for args in self._backend.search(search_string))

    def get_day_styles(self, day: dt.date, focus: bool) -> str | tuple[str, str] | None:
        calendars = self.get_calendars_on(day)
        if len(calendars) == 0:
            return None
        if self.color != "":
            return "highlight_days_color"
        if len(calendars) == 1:
            return "calendar " + calendars[0]
        if self.multiple != "" and not (self.multiple_on_overflow and len(calendars) == 2):
            return "highlight_days_multiple"
        return ("calendar " + calendars[0], "calendar " + calendars[1])

    def get_styles(self, date: dt.date, focus: bool) -> str | tuple[str, str] | None:
        if focus:
            if date == date.today():
                return "today focus"
            else:
                return "reveal focus"
        else:
            if date == date.today():
                return "today"
            else:
                if self.highlight_event_days:
                    return self.get_day_styles(date, focus)
                else:
                    return None



from collections.abc import Callable, Mapping
from importlib import metadata as importlib_metadata

# This is a shameless ripoff of mdformat's plugin extension API.
# see:
#   https://github.com/executablebooks/mdformat/blob/master/src/mdformat/plugins.py
#   https://setuptools.pypa.io/en/latest/userguide/entry_point.html


def _load_formatters() -> dict[str, Callable[[str], str]]:
    formatter_entrypoints = importlib_metadata.entry_points(group="khal.formatter")
    return {ep.name: ep.load() for ep in formatter_entrypoints}


FORMATTERS: Mapping[str, Callable[[str], str]] = _load_formatters()


def _load_color_themes() -> dict[str, list[tuple[str, ...]]]:
    color_theme_entrypoints = importlib_metadata.entry_points(group="khal.color_theme")
    return {ep.name: ep.load() for ep in color_theme_entrypoints}


THEMES: dict[
    str,
    list[tuple[str, ...]],
] = _load_color_themes()


def _load_commands() -> dict[str, Callable]:
    command_entrypoints = importlib_metadata.entry_points(group="khal.commands")
    return {ep.name: ep.load() for ep in command_entrypoints}


COMMANDS: dict[str, Callable] = _load_commands()



"""
Based off https://github.com/pimutils/python-vdir, which is itself based off
vdirsyncer.
"""

import contextlib
import errno
import os
import tempfile
import uuid
from collections.abc import Callable, Iterable
from hashlib import sha1
from typing import IO, Protocol

from khal.custom_types import PathLike, SupportsRaw


class HasMetaProtocol(Protocol):
    color_type: Callable

    def get_meta(self, key: str) -> str: ...

    def set_meta(self, key: str, value: str) -> None: ...


class cached_property:
    """A read-only @property that is only evaluated once. Only usable on class
    instances' methods.
    """

    def __init__(self, fget, doc=None) -> None:
        self.__name__ = fget.__name__
        self.__module__ = fget.__module__
        self.__doc__ = doc or fget.__doc__
        self.fget = fget

    def __get__(self, obj, cls):
        if obj is None:  # pragma: no cover
            return self
        obj.__dict__[self.__name__] = result = self.fget(obj)
        return result


SAFE_UID_CHARS = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_.-+@"


def _href_safe(uid: str, safe: str = SAFE_UID_CHARS) -> bool:
    return not bool(set(uid) - set(safe))


def _generate_href(uid: str | None = None, safe: str = SAFE_UID_CHARS) -> str:
    if not uid:
        return str(uuid.uuid4().hex)
    elif not _href_safe(uid, safe):
        return sha1(uid.encode()).hexdigest()
    else:
        return uid


def get_etag_from_file(f) -> str:
    """Get mtime-based etag from a filepath, file-like object or raw file
    descriptor.

    This function will flush/sync the file as much as necessary to obtain a
    correct mtime.
    """
    close_f = False
    if hasattr(f, "read"):
        f.flush()
        f = f.fileno()
    elif isinstance(f, str):
        flags = 0
        if os.path.isdir(f):
            flags = os.O_DIRECTORY
        f = os.open(f, flags)
        close_f = True

    # assure that all internal buffers associated with this file are
    # written to disk
    try:
        os.fsync(f)
        stat = os.fstat(f)
    finally:
        if close_f:
            os.close(f)

    mtime = getattr(stat, "st_mtime_ns", None)
    if mtime is None:
        mtime = stat.st_mtime
    return f"{mtime:.9f}"


class VdirError(IOError):
    def __init__(self, *args, **kwargs) -> None:
        for key, value in kwargs.items():
            if getattr(self, key, object()) not in [None, ""]:  # pragma: no cover
                raise TypeError(f"Invalid argument: {key}")
            setattr(self, key, value)

        super().__init__(*args)


class NotFoundError(VdirError):
    pass


class CollectionNotFoundError(VdirError):
    pass


class WrongEtagError(VdirError):
    pass


class AlreadyExistingError(VdirError):
    existing_href: str = ""


class Item:
    def __init__(self, raw: str) -> None:
        assert isinstance(raw, str)
        self.raw = raw

    @cached_property
    def uid(self) -> str | None:
        uid = ""
        lines = iter(self.raw.splitlines())
        for line in lines:
            if line.startswith("UID:"):
                uid += line[4:].strip()
                break

        for line in lines:
            if not line.startswith(" "):
                break
            uid += line[1:]

        return uid or None


@contextlib.contextmanager
def atomic_write(dest, overwrite=False):
    fd, src = tempfile.mkstemp(prefix=os.path.basename(dest), dir=os.path.dirname(dest))
    file = os.fdopen(fd, mode="wb")

    try:
        yield file
    except Exception:
        os.unlink(src)
        raise
    else:
        file.flush()
        file.close()

        if overwrite:
            os.rename(src, dest)
        else:
            os.link(src, dest)
            os.unlink(src)


class VdirBase:
    item_class = Item
    default_mode = 0o750

    def __init__(self, path: str, fileext: str, encoding: str = "utf-8") -> None:
        if not os.path.isdir(path):
            raise CollectionNotFoundError(path)
        self.path = path
        self.encoding = encoding
        self.fileext = fileext

    @classmethod
    def discover(cls, path: str, **kwargs) -> Iterable["VdirBase"]:
        try:
            collections = os.listdir(path)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
            return

        for collection in collections:
            collection_path = os.path.join(path, collection)
            if os.path.isdir(collection_path):
                yield cls(path=collection_path, **kwargs)

    @classmethod
    def create(cls, collection_name: PathLike, **kwargs: PathLike) -> dict[str, PathLike]:
        kwargs = dict(kwargs)
        path = kwargs["path"]

        pathn = os.path.join(path, collection_name)
        if not os.path.exists(pathn):
            os.makedirs(pathn, mode=cls.default_mode)
        elif not os.path.isdir(pathn):
            raise OSError(f"{pathn!r} is not a directory.")

        kwargs["path"] = pathn
        return kwargs

    def _get_filepath(self, href: str) -> str:
        return os.path.join(self.path, href)

    def _get_href(self, uid: str | None) -> str:
        return _generate_href(uid) + self.fileext

    def list(self) -> Iterable[tuple[str, str]]:
        for fname in os.listdir(self.path):
            fpath = os.path.join(self.path, fname)
            if os.path.isfile(fpath) and fname.endswith(self.fileext):
                yield fname, get_etag_from_file(fpath)

    def get(self, href: str) -> tuple[Item, str]:
        fpath = self._get_filepath(href)
        try:
            with open(fpath, "rb") as f:
                return (Item(f.read().decode(self.encoding)), get_etag_from_file(fpath))
        except OSError as e:
            if e.errno == errno.ENOENT:
                raise NotFoundError(href)
            else:
                raise

    def upload(self, item: SupportsRaw) -> tuple[str, str]:
        if not isinstance(item.raw, str):
            raise TypeError("item.raw must be a unicode string.")

        try:
            href = self._get_href(item.uid)
            _, etag = self._upload_impl(item, href)
        except OSError as e:
            if e.errno in (
                errno.ENAMETOOLONG,  # Unix
                errno.ENOENT,  # Windows
            ):
                # random href instead of UID-based
                href = self._get_href(None)
                _, etag = self._upload_impl(item, href)
            else:
                raise
        return href, etag

    def _upload_impl(self, item: SupportsRaw, href: str) -> tuple[str, str]:
        fpath = self._get_filepath(href)
        try:
            f: IO
            with atomic_write(fpath, overwrite=False) as f:
                f.write(item.raw.encode(self.encoding))
                return fpath, get_etag_from_file(f)
        except OSError as e:
            if e.errno == errno.EEXIST:
                raise AlreadyExistingError(existing_href=href)
            else:
                raise

    def update(self, href: str, item: SupportsRaw, etag: str) -> str:
        fpath = self._get_filepath(href)
        if not os.path.exists(fpath):
            raise NotFoundError(item.uid)
        actual_etag = get_etag_from_file(fpath)
        if etag != actual_etag:
            raise WrongEtagError(etag, actual_etag)

        if not isinstance(item.raw, str):
            raise TypeError("item.raw must be a unicode string.")

        with atomic_write(fpath, overwrite=True) as f:
            f.write(item.raw.encode(self.encoding))
            etag = get_etag_from_file(f)

        return etag

    def delete(self, href: str, etag: str | None) -> None:
        fpath = self._get_filepath(href)
        if not os.path.isfile(fpath):
            raise NotFoundError(href)
        actual_etag = get_etag_from_file(fpath)
        if etag != actual_etag:
            raise WrongEtagError(etag, actual_etag)
        os.remove(fpath)

    def get_meta(self, key: str) -> str | None:
        fpath = os.path.join(self.path, key)
        try:
            with open(fpath, "rb") as f:
                return f.read().decode(self.encoding).strip() or None
        except OSError as e:
            if e.errno == errno.ENOENT:
                return None
            else:
                raise

    def set_meta(self, key: str, value: str) -> None:
        value = value or ""
        assert isinstance(value, str)
        fpath = os.path.join(self.path, key)
        with atomic_write(fpath, overwrite=True) as f:
            f.write(value.encode(self.encoding))


class Color:
    def __init__(self, x: str) -> None:
        if not x:
            raise ValueError("Color is false-ish.")
        if not x.startswith("#"):
            raise ValueError("Color must start with a #.")
        if len(x) != 7:
            raise ValueError("Color must not have shortcuts. #ffffff instead of #fff")
        self.raw: str = x.upper()

    @cached_property
    def rgb(self) -> tuple[int, int, int]:
        x = self.raw

        r = x[1:3]
        g = x[3:5]
        b = x[5:8]

        if len(r) == len(g) == len(b) == 2:
            return int(r, 16), int(g, 16), int(b, 16)
        else:
            raise ValueError(f"Unable to parse color value: {self.raw}")


class ColorMixin:
    color_type: type[Color] = Color

    def get_color(self: HasMetaProtocol) -> str | None:
        try:
            return self.color_type(self.get_meta("color"))
        except ValueError:
            return None

    def set_color(self: HasMetaProtocol, value: str) -> None:
        self.set_meta("color", self.color_type(value).raw)


class DisplayNameMixin:
    def get_displayname(self: HasMetaProtocol) -> str:
        return self.get_meta("displayname")

    def set_displayname(self: HasMetaProtocol, value: str) -> None:
        self.set_meta("displayname", value)


class Vdir(VdirBase, ColorMixin, DisplayNameMixin):
    pass



import datetime as dt
import os
from typing import Literal, Protocol, TypedDict

import pytz


class CalendarConfiguration(TypedDict):
    name: str
    path: str
    readonly: bool
    color: str
    priority: int
    ctype: str
    addresses: str


class LocaleConfiguration(TypedDict):
    local_timezone: pytz.BaseTzInfo
    default_timezone: pytz.BaseTzInfo
    timeformat: str
    dateformat: str
    longdateformat: str
    datetimeformat: str
    longdatetimeformat: str
    weeknumbers: str | bool
    firstweekday: int
    unicode_symbols: bool


class SupportsRaw(Protocol):
    @property
    def uid(self) -> str | None: ...

    @property
    def raw(self) -> str: ...


# set this to TypeAlias once we support that python version (PEP613)
EventTuple = tuple[
    str,
    str,
    dt.date | dt.datetime,
    dt.date | dt.datetime,
    str,
    str,
    str,
]


# Only need for RRuleMapType
class RRuleMapBase(TypedDict):
    freq: str


class RRuleMapType(RRuleMapBase, total=False):
    # not required keys go in here
    # TODO remove if either `NotRequired` is supported by mypy or the oldest
    # python we support is 3.11 (see PEP 655)
    until: dt.datetime


class EventCreationTypes(TypedDict):
    dtstart: dt.date
    dtend: dt.date
    summary: str
    description: str
    allday: bool
    location: str | None
    categories: str | list[str] | None
    repeat: str | None
    until: str
    alarms: str
    timezone: pytz.BaseTzInfo
    url: str


PathLike = str | os.PathLike

WeekNumbersType = Literal["left", "right", False]
MonthDisplayType = Literal["firstday", "firstfullweek"]
